=== modified file 'NEWS'
--- NEWS	2005-09-29 20:34:25 +0000
+++ NEWS	2005-10-03 01:42:16 +0000
@@ -1,5 +1,6 @@
 DEVELOPMENT HEAD
 
+
   BUG FIXES:
   
     * bzr-man.py has been updated to work again. Contributed by
@@ -8,6 +9,7 @@
     * Locking is now done with fcntl.lockf which works with NFS
       file systems. Contributed by Harald Meland.
 
+
   IMPROVEMENTS:
 
     * 'branch' now accepts a --basis parameter which will take advantage
@@ -18,6 +20,13 @@
       Existing branches can be upgraded to this format using 
       'bzr upgrade'.
 
+    * Symlinks are now versionable. Initial patch by 
+      Erik Toubro Nielsen, updated to head by Robert Collins.
+
+    * Executable bits are tracked on files. Patch from Gustavo
+      Niemeyer.
+
+
   INTERNALS:
 
     * Test case failures have the exception printed after the log 

=== modified file 'bzrlib/add.py'
--- bzrlib/add.py	2005-09-29 20:34:25 +0000
+++ bzrlib/add.py	2005-10-02 21:51:29 +0000
@@ -95,7 +95,7 @@
 
         kind = bzrlib.osutils.file_kind(af)
 
-        if kind != 'file' and kind != 'directory':
+        if kind != 'file' and kind != 'directory' and kind != 'symlink':
             if f in user_list:
                 raise BadFileKindError("cannot add %s of type %s" % (f, kind))
             else:

=== modified file 'bzrlib/annotate.py'
--- bzrlib/annotate.py	2005-09-22 13:32:02 +0000
+++ bzrlib/annotate.py	2005-09-30 14:27:41 +0000
@@ -51,5 +51,5 @@
     rp = b.relpath(sys.argv[1])
     tree = b.revision_tree(b.last_revision())
     file_id = tree.inventory.path2id(rp)
-    file_version = tree.inventory[file_id].text_version
+    file_version = tree.inventory[file_id].revision
     annotate_file(b, file_version, file_id, sys.stdout)

=== modified file 'bzrlib/branch.py'
--- bzrlib/branch.py	2005-09-30 07:38:09 +0000
+++ bzrlib/branch.py	2005-10-02 21:51:29 +0000
@@ -520,11 +520,9 @@
         
         mutter('wrote working inventory')
             
-
     inventory = property(read_working_inventory, _write_inventory, None,
                          """Inventory for the working copy.""")
 
-
     def add(self, files, ids=None):
         """Make files versioned.
 
@@ -578,10 +576,10 @@
                     kind = file_kind(fullpath)
                 except OSError:
                     # maybe something better?
-                    raise BzrError('cannot add: not a regular file or directory: %s' % quotefn(f))
+                    raise BzrError('cannot add: not a regular file, symlink or directory: %s' % quotefn(f))
 
-                if kind != 'file' and kind != 'directory':
-                    raise BzrError('cannot add: not a regular file or directory: %s' % quotefn(f))
+                if kind not in ('file', 'directory', 'symlink'):
+                    raise BzrError('cannot add: not a regular file, symlink or directory: %s' % quotefn(f))
 
                 if file_id is None:
                     file_id = gen_file_id(f)
@@ -652,7 +650,6 @@
         finally:
             self.unlock()
 
-
     # FIXME: this doesn't need to be a branch method
     def set_inventory(self, new_inventory_list):
         from bzrlib.inventory import Inventory, InventoryEntry
@@ -664,7 +661,6 @@
             inv.add(InventoryEntry(file_id, name, kind, parent))
         self._write_inventory(inv)
 
-
     def unknowns(self):
         """Return all unknown files.
 
@@ -695,7 +691,6 @@
         finally:
             self.unlock()
 
-
     def has_revision(self, revision_id):
         """True if this branch has a copy of the revision.
 
@@ -704,7 +699,6 @@
         return (revision_id is None
                 or revision_id in self.revision_store)
 
-
     def get_revision_xml_file(self, revision_id):
         """Return XML file object for revision object."""
         if not revision_id or not isinstance(revision_id, basestring):
@@ -719,6 +713,8 @@
         finally:
             self.unlock()
 
+    #deprecated
+    get_revision_xml = get_revision_xml_file
 
     def get_revision_xml(self, revision_id):
         return self.get_revision_xml_file(revision_id).read()
@@ -738,7 +734,6 @@
         assert r.revision_id == revision_id
         return r
 
-
     def get_revision_delta(self, revno):
         """Return the delta for one revision.
 
@@ -762,12 +757,16 @@
 
     def get_revision_sha1(self, revision_id):
         """Hash the stored value of a revision, and return it."""
+        # In the future, revision entries will be signed. At that
+        # point, it is probably best *not* to include the signature
+        # in the revision hash. Because that lets you re-sign
+        # the revision, (add signatures/remove signatures) and still
+        # have all hash pointers stay consistent.
+        # But for now, just hash the contents.
         return bzrlib.osutils.sha_file(self.get_revision_xml_file(revision_id))
 
-
     def _get_ancestry_weave(self):
         return self.control_weaves.get_weave('ancestry')
-        
 
     def get_ancestry(self, revision_id):
         """Return a list of revision-ids integrated by a revision.
@@ -778,17 +777,14 @@
         w = self._get_ancestry_weave()
         return [None] + [l[:-1] for l in w.get_iter(w.lookup(revision_id))]
 
-
     def get_inventory_weave(self):
         return self.control_weaves.get_weave('inventory')
 
-
     def get_inventory(self, revision_id):
         """Get Inventory object by hash."""
         xml = self.get_inventory_xml(revision_id)
         return bzrlib.xml5.serializer_v5.read_inventory_from_string(xml)
 
-
     def get_inventory_xml(self, revision_id):
         """Get inventory XML as a file object."""
         try:
@@ -798,13 +794,11 @@
         except IndexError:
             raise bzrlib.errors.HistoryMissing(self, 'inventory', revision_id)
 
-
     def get_inventory_sha1(self, revision_id):
         """Return the sha1 hash of the inventory entry
         """
         return self.get_revision(revision_id).inventory_sha1
 
-
     def get_revision_inventory(self, revision_id):
         """Return inventory of a past revision."""
         # TODO: Unify this with get_inventory()
@@ -815,7 +809,6 @@
         else:
             return self.get_inventory(revision_id)
 
-
     def revision_history(self):
         """Return sequence of revision hashes on to this branch."""
         self.lock_read()
@@ -825,7 +818,6 @@
         finally:
             self.unlock()
 
-
     def common_ancestor(self, other, self_revno=None, other_revno=None):
         """
         >>> from bzrlib.commit import commit
@@ -1315,8 +1307,6 @@
         copytree(self.base, base, symlinks=True)
         return ScratchBranch(base=base)
 
-
-        
     def __del__(self):
         self.destroy()
 

=== modified file 'bzrlib/builtins.py'
--- bzrlib/builtins.py	2005-09-30 06:39:42 +0000
+++ bzrlib/builtins.py	2005-10-03 01:42:16 +0000
@@ -70,7 +70,7 @@
     # relative to a revision, or between revisions
 
     takes_args = ['file*']
-    takes_options = ['all', 'show-ids', 'revision']
+    takes_options = ['all', 'show-ids']
     aliases = ['st', 'stat']
     
     def run(self, all=False, show_ids=False, file_list=None, revision=None):
@@ -556,7 +556,6 @@
         bzr commit -m 'imported project'
     """
     def run(self):
-        from bzrlib.branch import Branch
         Branch.initialize('.')
 
 
@@ -652,7 +651,7 @@
         b = Branch.open_containing('.')
         td = compare_trees(b.basis_tree(), b.working_tree())
 
-        for path, id, kind in td.modified:
+        for path, id, kind, text_modified, meta_modified in td.modified:
             print path
 
 
@@ -1247,7 +1246,6 @@
                 if None in revision:
                     raise BzrCommandError(
                         "Merge doesn't permit that revision specifier.")
-                from bzrlib.branch import Branch
                 b = Branch.open(branch)
 
                 base = [branch, revision[0].in_history(b).revno]
@@ -1278,7 +1276,6 @@
 
     def run(self, revision=None, no_backup=False, file_list=None):
         from bzrlib.merge import merge
-        from bzrlib.branch import Branch
         from bzrlib.commands import parse_spec
 
         if file_list is not None:

=== modified file 'bzrlib/changeset.py'
--- bzrlib/changeset.py	2005-09-26 08:56:15 +0000
+++ bzrlib/changeset.py	2005-10-03 01:42:16 +0000
@@ -19,6 +19,7 @@
 import stat
 from bzrlib.trace import mutter
 from bzrlib.osutils import rename
+import bzrlib
 
 # XXX: mbp: I'm not totally convinced that we should handle conflicts
 # as part of changeset application, rather than only in the merge
@@ -862,7 +863,7 @@
             continue
         new_path = os.path.join(dir, new_tree_path)
         old_path = changed_inventory.get(entry.id)
-        if os.path.exists(new_path):
+        if bzrlib.osutils.lexists(new_path):
             if conflict_handler.target_exists(entry, new_path, old_path) == \
                 "skip":
                 continue
@@ -1383,18 +1384,20 @@
 
         if cs_entry is None:
             return None
+
+        full_path_a = self.tree_a.readonly_path(id)
+        full_path_b = self.tree_b.readonly_path(id)
+        stat_a = self.lstat(full_path_a)
+        stat_b = self.lstat(full_path_b)
+
+        cs_entry.metadata_change = self.make_mode_change(stat_a, stat_b)
+
         if id in self.tree_a and id in self.tree_b:
             a_sha1 = self.tree_a.get_file_sha1(id)
             b_sha1 = self.tree_b.get_file_sha1(id)
             if None not in (a_sha1, b_sha1) and a_sha1 == b_sha1:
                 return cs_entry
 
-        full_path_a = self.tree_a.readonly_path(id)
-        full_path_b = self.tree_b.readonly_path(id)
-        stat_a = self.lstat(full_path_a)
-        stat_b = self.lstat(full_path_b)
-        
-        cs_entry.metadata_change = self.make_mode_change(stat_a, stat_b)
         cs_entry.contents_change = self.make_contents_change(full_path_a,
                                                              stat_a, 
                                                              full_path_b, 

=== modified file 'bzrlib/check.py'
--- bzrlib/check.py	2005-09-27 07:24:40 +0000
+++ bzrlib/check.py	2005-09-30 02:54:51 +0000
@@ -28,7 +28,7 @@
 
 import bzrlib.ui
 from bzrlib.trace import note, warning
-from bzrlib.osutils import rename, sha_string, fingerprint_file, sha_strings
+from bzrlib.osutils import rename, sha_string, fingerprint_file
 from bzrlib.trace import mutter
 from bzrlib.errors import BzrCheckError, NoSuchRevision
 from bzrlib.inventory import ROOT_ID
@@ -148,7 +148,8 @@
                                     % (file_id, rev_id))
             seen_ids[file_id] = True
         for file_id in inv:
-            self._check_one_entry(rev_id, inv, tree, file_id)
+            ie = inv[file_id]
+            ie.check(self, rev_id, inv, tree)
         seen_names = {}
         for path, ie in inv.iter_entries():
             if path in seen_names:
@@ -157,42 +158,6 @@
                                     % (path, rev_id))
             seen_names[path] = True
 
-        
-    def _check_one_entry(self, rev_id, inv, tree, file_id):
-        ie = inv[file_id]
-        if ie.parent_id != None:
-            if not inv.has_id(ie.parent_id):
-                raise BzrCheckError('missing parent {%s} in inventory for revision {%s}'
-                        % (ie.parent_id, rev_id))
-        if ie.kind == 'file':
-            text_version = ie.text_version
-            t = (file_id, text_version)
-            if t in self.checked_texts:
-                prev_sha = self.checked_texts[t] 
-                if prev_sha != ie.text_sha1:
-                    raise BzrCheckError('mismatched sha1 on {%s} in {%s}' %
-                                        (file_id, rev_id))
-                else:
-                    self.repeated_text_cnt += 1
-                    return
-            mutter('check version {%s} of {%s}', rev_id, file_id)
-            file_lines = tree.get_file_lines(file_id)
-            self.checked_text_cnt += 1 
-            if ie.text_size != sum(map(len, file_lines)):
-                raise BzrCheckError('text {%s} wrong size' % ie.text_id)
-            if ie.text_sha1 != sha_strings(file_lines):
-                raise BzrCheckError('text {%s} wrong sha1' % ie.text_id)
-            self.checked_texts[t] = ie.text_sha1
-        elif ie.kind == 'directory':
-            if ie.text_sha1 != None or ie.text_size != None or ie.text_id != None:
-                raise BzrCheckError('directory {%s} has text in revision {%s}'
-                        % (file_id, rev_id))
-        elif ie.kind == 'root_directory':
-            pass
-        else:
-            raise BzrCheckError('unknown entry kind %r in revision {%s}' % 
-                                (ie.kind, rev_id))
-
 
 def check(branch):
     """Run consistency checks on a branch."""

=== modified file 'bzrlib/clone.py'
--- bzrlib/clone.py	2005-09-30 07:38:34 +0000
+++ bzrlib/clone.py	2005-10-02 21:51:29 +0000
@@ -47,6 +47,7 @@
 import os
 import sys
 
+import bzrlib
 from bzrlib.merge import build_working_dir
 from bzrlib.branch import Branch
 from bzrlib.trace import mutter, note
@@ -75,7 +76,7 @@
     if basis_branch is not None:
         note("basis_branch is not supported for fast weave copy yet.")
     history = _get_truncated_history(branch_from, revision)
-    if not os.path.exists(to_location):
+    if not bzrlib.osutils.lexists(to_location):
         os.mkdir(to_location)
     branch_to = Branch.initialize(to_location)
     mutter("copy branch from %s to %s", branch_from, branch_to)
@@ -140,7 +141,7 @@
     """
     assert isinstance(branch_from, Branch)
     assert isinstance(to_location, basestring)
-    if not os.path.exists(to_location):
+    if not bzrlib.osutils.lexists(to_location):
         os.mkdir(to_location)
     br_to = Branch.initialize(to_location)
     mutter("copy branch from %s to %s", branch_from, br_to)

=== modified file 'bzrlib/commands.py'
--- bzrlib/commands.py	2005-09-29 20:34:25 +0000
+++ bzrlib/commands.py	2005-10-02 21:51:29 +0000
@@ -32,6 +32,13 @@
 # would help with validation and shell completion.
 
 
+# TODO: Help messages for options.
+
+# TODO: Define arguments by objects, rather than just using names.
+# Those objects can specify the expected type of the argument, which
+# would help with validation and shell completion.
+
+
 
 import sys
 import os

=== modified file 'bzrlib/commit.py'
--- bzrlib/commit.py	2005-09-30 06:39:42 +0000
+++ bzrlib/commit.py	2005-10-02 22:47:02 +0000
@@ -214,10 +214,9 @@
             self._check_parents_present()
             
             self._remove_deleted()
-            self.new_inv = Inventory()
-            self._store_entries()
+            self._populate_new_inv()
+            self._store_snapshot()
             self._report_deletes()
-            self._set_name_versions()
 
             if not (self.allow_pointless
                     or len(self.parents) > 1
@@ -234,15 +233,13 @@
         finally:
             self.branch.unlock()
 
-
-
     def _record_inventory(self):
         """Store the inventory for the new revision."""
         inv_text = serializer_v5.write_inventory_to_string(self.new_inv)
         self.inv_sha1 = sha_string(inv_text)
         s = self.branch.control_weaves
         s.add_text('inventory', self.rev_id,
-                   split_lines(inv_text), self.parents)
+                   split_lines(inv_text), self.present_parents)
 
     def _escape_commit_message(self):
         """Replace xml-incompatible control characters."""
@@ -271,39 +268,42 @@
         s = self.branch.control_weaves
         w = s.get_weave_or_empty('ancestry')
         lines = self._make_ancestry(w)
-        w.add(self.rev_id, self.parents, lines)
+        w.add(self.rev_id, self.present_parents, lines)
         s.put_weave('ancestry', w)
 
-
     def _make_ancestry(self, ancestry_weave):
         """Return merged ancestry lines.
 
         The lines are revision-ids followed by newlines."""
-        parent_ancestries = [ancestry_weave.get(p) for p in self.parents]
+        parent_ancestries = [ancestry_weave.get(p) for p in self.present_parents]
         new_lines = merge_ancestry_lines(self.rev_id, parent_ancestries)
         mutter('merged ancestry of {%s}:\n%s', self.rev_id, ''.join(new_lines))
         return new_lines
 
-
     def _gather_parents(self):
+        """Record the parents of a merge for merge detection."""
         pending_merges = self.branch.pending_merges()
         self.parents = []
         self.parent_trees = []
+        self.present_parents = []
         precursor_id = self.branch.last_revision()
         if precursor_id:
             self.parents.append(precursor_id)
-            self.parent_trees.append(self.basis_tree)
         self.parents += pending_merges
-        self.parent_trees.extend(map(self.branch.revision_tree, pending_merges))
-
+        for revision in self.parents:
+            if self.branch.has_revision(revision):
+                self.parent_trees.append(self.branch.revision_tree(revision))
+                self.present_parents.append(revision)
 
     def _check_parents_present(self):
         for parent_id in self.parents:
             mutter('commit parent revision {%s}', parent_id)
             if not self.branch.has_revision(parent_id):
-                warning("can't commit a merge from an absent parent")
-                raise HistoryMissing(self.branch, 'revision', parent_id)
-
+                if parent_id == self.branch.last_revision():
+                    warning("parent is pissing %r", parent_id)
+                    raise HistoryMissing(self.branch, 'revision', parent_id)
+                else:
+                    mutter("commit will ghost revision %r", parent_id)
             
     def _make_revision(self):
         """Record a new revision object for this commit."""
@@ -347,7 +347,7 @@
             self.branch._write_inventory(self.work_inv)
 
 
-    def _find_file_parents(self, file_id):
+    def _find_entry_parents(self, file_id):
         """Return the text versions and hashes for all file parents.
 
         Returned as a map from text version to inventory entry.
@@ -359,77 +359,54 @@
         for tree in self.parent_trees:
             if file_id in tree.inventory:
                 ie = tree.inventory[file_id]
-                assert ie.kind == 'file'
                 assert ie.file_id == file_id
-                if ie.text_version in r:
-                    assert r[ie.text_version] == ie
+                if ie.revision in r:
+                    assert r[ie.revision] == ie
                 else:
-                    r[ie.text_version] = ie
+                    r[ie.revision] = ie
         return r
 
-
-    def _set_name_versions(self):
-        """Pass over inventory and mark new entry version as needed.
-
-        Files get a new name version when they are new, have a
-        different parent, or a different name from in the
-        basis inventory, or if the file is in a different place
-        to any of the parents."""
+    def _store_snapshot(self):
+        """Pass over inventory and record a snapshot.
+
+        Entries get a new revision when they are modified in 
+        any way, which includes a merge with a new set of
+        parents that have the same entry. Currently we do not
+        check for that set being ancestors of each other - and
+        we should - only parallel children should count for this
+        test see find_entry_parents to correct this. FIXME <---
+        I.e. if we are merging in revision FOO, and our
+        copy of file id BAR is identical to FOO.BAR, we should
+        generate a new revision of BAR IF and only IF FOO is
+        neither a child of our current tip, nor an ancestor of
+        our tip. The presence of FOO in our store should not 
+        affect this logic UNLESS we are doing a merge of FOO,
+        or a child of FOO.
+        """
         # XXX: Need to think more here about when the user has
         # made a specific decision on a particular value -- c.f.
         # mark-merge.  
         for path, ie in self.new_inv.iter_entries():
-            old_version = None
-            file_id = ie.file_id
-            for parent_tree in self.parent_trees:
-                parent_inv = parent_tree.inventory
-                if file_id not in parent_inv:
-                    continue
-                parent_ie = parent_inv[file_id]
-                if parent_ie.parent_id != ie.parent_id:
-                    old_version = None
-                    break
-                elif parent_ie.name != ie.name:
-                    old_version = None
-                    break
-                elif old_version is None:
-                    old_version = parent_ie.name_version
-                elif old_version != parent_ie.name_version:
-                    old_version = None
-                    break
-                else:
-                    pass                # so far so good
-            if old_version is None:
-                mutter('new name_version for {%s}', file_id)
-                ie.name_version = self.rev_id
+            previous_entries = self._find_entry_parents(ie. file_id)
+            if ie.revision is None:
+                change = ie.snapshot(self.rev_id, path, previous_entries,
+                                     self.work_tree, self.weave_store)
             else:
-                mutter('name_version for {%s} inherited as {%s}',
-                       file_id, old_version)
-                ie.name_version = old_version
-
-
-    def _store_entries(self):
-        """Build revision inventory and store modified files.
-
-        This is called with new_inv a new empty inventory.  Depending on
-        which files are selected for commit, and which ones have
-        been modified or merged, new inventory entries are built
-        based on the working and parent inventories.
-
-        As a side-effect this stores new text versions for committed
-        files with text changes or merges.
-
-        Each entry can have one of several things happen:
-
-        carry_file -- carried from the previous version (if not
-            selected for commit)
-
-        commit_nonfile -- no text to worry about
-
-        commit_old_text -- same text, may have moved
-
-        commit_file -- new text version
+                change = "unchanged"
+            note("%s %s", change, path)
+
+    def _populate_new_inv(self):
+        """Build revision inventory.
+
+        This creates a new empty inventory. Depending on
+        which files are selected for commit, and what is present in the
+        current tree, the new inventory is populated. inventory entries 
+        which are candidates for modification have their revision set to
+        None; inventory entries that are carried over untouched have their
+        revision set to their prior value.
         """
+        mutter("Selecting files for commit with filter %s", self.specific_files)
+        self.new_inv = Inventory()
         for path, new_ie in self.work_inv.iter_entries():
             file_id = new_ie.file_id
             mutter('check %s {%s}', path, new_ie.file_id)
@@ -438,69 +415,22 @@
                     mutter('%s not selected for commit', path)
                     self._carry_file(file_id)
                     continue
-            if new_ie.kind != 'file':
-                self._commit_nonfile(file_id)
-                continue
-            
-            file_parents = self._find_file_parents(file_id)
-            mutter('parents of %s are %r', path, file_parents)
-            if len(file_parents) == 1:
-                parent_ie = file_parents.values()[0]
-                wc_sha1 = self.work_tree.get_file_sha1(file_id)
-                if parent_ie.text_sha1 == wc_sha1:
-                    # text not changed or merged
-                    self._commit_old_text(file_id, parent_ie)
-                    continue
-            # file is either new, or a file merge; need to record
-            # a new version
-            if len(file_parents) > 1:
-                note('merged %s', path)
-            elif len(file_parents) == 0:
-                note('added %s', path)
-            else:
-                note('modified %s', path)
-            self._commit_file(new_ie, file_id, file_parents)
-
-
-    def _commit_nonfile(self, file_id):
-        self.new_inv.add(self.work_inv[file_id].copy())
-
+            mutter('%s selected for commit', path)
+            ie = new_ie.copy()
+            ie.revision = None
+            self.new_inv.add(ie)
 
     def _carry_file(self, file_id):
         """Carry the file unchanged from the basis revision."""
         if self.basis_inv.has_id(file_id):
             self.new_inv.add(self.basis_inv[file_id].copy())
 
-
-    def _commit_old_text(self, file_id, parent_ie):
-        """Keep the same text as last time, but possibly a different name."""
-        ie = self.work_inv[file_id].copy()
-        ie.text_version = parent_ie.text_version
-        ie.text_size = parent_ie.text_size
-        ie.text_sha1 = parent_ie.text_sha1
-        self.new_inv.add(ie)
-
-
     def _report_deletes(self):
         for file_id in self.basis_inv:
             if file_id not in self.new_inv:
                 note('deleted %s', self.basis_inv.id2path(file_id))
 
 
-    def _commit_file(self, new_ie, file_id, file_parents):                    
-        mutter('store new text for {%s} in revision {%s}',
-               file_id, self.rev_id)
-        new_lines = self.work_tree.get_file(file_id).readlines()
-        self._add_text_to_weave(file_id, new_lines, file_parents)
-        new_ie.text_version = self.rev_id
-        new_ie.text_sha1 = sha_strings(new_lines)
-        new_ie.text_size = sum(map(len, new_lines))
-        self.new_inv.add(new_ie)
-
-
-    def _add_text_to_weave(self, file_id, new_lines, parents):
-        self.weave_store.add_text(file_id, self.rev_id, new_lines, parents)
-
 
 def _gen_revision_id(branch, when):
     """Return new revision-id."""

=== modified file 'bzrlib/delta.py'
--- bzrlib/delta.py	2005-09-27 07:24:40 +0000
+++ bzrlib/delta.py	2005-10-03 01:42:16 +0000
@@ -26,16 +26,18 @@
     removed
         (path, id, kind)
     renamed
-        (oldpath, newpath, id, kind, text_modified)
+        (oldpath, newpath, id, kind, text_modified, meta_modified)
     modified
-        (path, id, kind)
+        (path, id, kind, text_modified, meta_modified)
     unchanged
         (path, id, kind)
 
     Each id is listed only once.
 
     Files that are both modified and renamed are listed only in
-    renamed, with the text_modified flag true.
+    renamed, with the text_modified flag true. The text_modified
+    applies either to the the content of the file or the target of the
+    symbolic link, depending of the kind of file.
 
     Files are only considered renamed if their name has changed or
     their parent directory has changed.  Renaming a directory
@@ -87,12 +89,17 @@
 
     def show(self, to_file, show_ids=False, show_unchanged=False):
         def show_list(files):
-            for path, fid, kind in files:
+            for item in files:
+                path, fid, kind = item[:3]
+
                 if kind == 'directory':
                     path += '/'
                 elif kind == 'symlink':
                     path += '@'
-                    
+
+                if len(item) == 5 and item[4]:
+                    path += '*'
+
                 if show_ids:
                     print >>to_file, '  %-30s %s' % (path, fid)
                 else:
@@ -108,7 +115,10 @@
 
         if self.renamed:
             print >>to_file, 'renamed:'
-            for oldpath, newpath, fid, kind, text_modified in self.renamed:
+            for (oldpath, newpath, fid, kind,
+                 text_modified, meta_modified) in self.renamed:
+                if meta_modified:
+                    newpath += '*'
                 if show_ids:
                     print >>to_file, '  %s => %s %s' % (oldpath, newpath, fid)
                 else:
@@ -176,9 +186,24 @@
                 old_sha1 = old_tree.get_file_sha1(file_id)
                 new_sha1 = new_tree.get_file_sha1(file_id)
                 text_modified = (old_sha1 != new_sha1)
+                old_exec = old_tree.is_executable(file_id)
+                new_exec = new_tree.is_executable(file_id)
+                meta_modified = (old_exec != new_exec)
+            elif kind == 'symlink':
+                t1 = old_tree.get_symlink_target(file_id)
+                t2 = new_tree.get_symlink_target(file_id)
+                if t1 != t2:
+                    mutter("    symlink target changed")
+                    # FIXME: which should we use ?
+                    text_modified = True
+                    meta_modified = False
+                else:
+                    text_modified = False
+                    meta_modified = False
             else:
                 ## mutter("no text to check for %r %r" % (file_id, kind))
                 text_modified = False
+                meta_modified = False
 
             # TODO: Can possibly avoid calculating path strings if the
             # two files are unchanged and their names and parents are
@@ -190,9 +215,10 @@
                 delta.renamed.append((old_inv.id2path(file_id),
                                       new_inv.id2path(file_id),
                                       file_id, kind,
-                                      text_modified))
-            elif text_modified:
-                delta.modified.append((new_inv.id2path(file_id), file_id, kind))
+                                      text_modified, meta_modified))
+            elif text_modified or meta_modified:
+                delta.modified.append((new_inv.id2path(file_id), file_id, kind,
+                                       text_modified, meta_modified))
             elif want_unchanged:
                 delta.unchanged.append((new_inv.id2path(file_id), file_id, kind))
         else:

=== modified file 'bzrlib/diff.py'
--- bzrlib/diff.py	2005-09-29 20:34:25 +0000
+++ bzrlib/diff.py	2005-10-03 01:42:16 +0000
@@ -62,8 +62,6 @@
     print >>to_file
 
 
-
-
 def external_diff(old_label, oldlines, new_label, newlines, to_file,
                   diff_opts):
     """Display a diff by calling out to the external diff program."""
@@ -141,8 +139,6 @@
     finally:
         oldtmpf.close()                 # and delete
         newtmpf.close()
-    
-
 
 def show_diff(b, revision, specific_files, external_diff_options=None,
               revision2=None, output=None):
@@ -219,7 +215,6 @@
                       DEVNULL, 
                       [],
                       to_file)
-
     for path, file_id, kind in delta.added:
         print >>to_file, '=== added %s %r' % (kind, path)
         if kind == 'file':
@@ -228,26 +223,44 @@
                       new_label + path,
                       new_tree.get_file(file_id).readlines(),
                       to_file)
-
-    for old_path, new_path, file_id, kind, text_modified in delta.renamed:
-        print >>to_file, '=== renamed %s %r => %r' % (kind, old_path, new_path)
+    for (old_path, new_path, file_id, kind,
+         text_modified, meta_modified) in delta.renamed:
+        prop_str = get_prop_change(meta_modified)
+        print >>to_file, '=== renamed %s %r => %r%s' % (
+                          kind, old_path, new_path, prop_str)
+        _maybe_diff_file_or_symlink(old_label, old_path, old_tree, file_id,
+                                    new_label, new_path, new_tree,
+                                    text_modified, kind, to_file, diff_file)
+    for path, file_id, kind, text_modified, meta_modified in delta.modified:
+        prop_str = get_prop_change(meta_modified)
+        print >>to_file, '=== modified %s %r%s' % (kind, path, prop_str)
         if text_modified:
+            _maybe_diff_file_or_symlink(old_label, path, old_tree, file_id,
+                                        new_label, path, new_tree,
+                                        True, kind, to_file, diff_file)
+    
+
+def get_prop_change(meta_modified):
+    if meta_modified:
+        return " (properties changed)"
+    else:
+        return  ""
+
+
+def _maybe_diff_file_or_symlink(old_label, old_path, old_tree, file_id,
+                                new_label, new_path, new_tree, text_modified,
+                                kind, to_file, diff_file):
+    if text_modified:
+        if kind == 'file':
             diff_file(old_label + old_path,
                       old_tree.get_file(file_id).readlines(),
                       new_label + new_path,
                       new_tree.get_file(file_id).readlines(),
                       to_file)
-
-    for path, file_id, kind in delta.modified:
-        print >>to_file, '=== modified %s %r' % (kind, path)
-        if kind == 'file':
-            diff_file(old_label + path,
-                      old_tree.get_file(file_id).readlines(),
-                      new_label + path,
-                      new_tree.get_file(file_id).readlines(),
-                      to_file)
-
-
-
-
-
+        elif kind == 'symlink':
+            _diff_symlink(old_tree, new_tree, file_id, to_file)
+            
+def _diff_symlink(old_tree, new_tree, file_id, to_file):
+    t1 = old_tree.get_symlink_target(file_id)
+    t2 = new_tree.get_symlink_target(file_id)
+    print >>to_file, '=== target changed %r => %r' % (t1, t2)

=== modified file 'bzrlib/errors.py'
--- bzrlib/errors.py	2005-09-29 20:34:25 +0000
+++ bzrlib/errors.py	2005-10-02 21:51:29 +0000
@@ -138,6 +138,15 @@
         self.not_ancestor_id = not_ancestor_id
 
 
+class NotAncestor(BzrError):
+    def __init__(self, rev_id, not_ancestor_id):
+        self.rev_id = rev_id
+        self.not_ancestor_id = not_ancestor_id
+        msg = "Revision %s is not an ancestor of %s" % (not_ancestor_id, 
+                                                        rev_id)
+        BzrError.__init__(self, msg)
+
+
 class InstallFailed(BzrError):
     def __init__(self, revisions):
         msg = "Could not install revisions:\n%s" % " ,".join(revisions)

=== modified file 'bzrlib/fetch.py'
--- bzrlib/fetch.py	2005-09-30 05:15:17 +0000
+++ bzrlib/fetch.py	2005-10-02 23:17:00 +0000
@@ -184,8 +184,12 @@
                rev.committer,
                len(rev.parent_ids))
         self._copy_new_texts(rev_id, inv)
-        self._copy_inventory(rev_id, inv_xml, rev.parent_ids)
-        self._copy_ancestry(rev_id, rev.parent_ids)
+        parents = rev.parent_ids
+        for parent in parents:
+            if not self.to_branch.has_revision(parent):
+                parents.pop(parents.index(parent))
+        self._copy_inventory(rev_id, inv_xml, parents)
+        self._copy_ancestry(rev_id, parents)
         self.to_branch.revision_store.add(StringIO(rev_xml), rev_id)
 
 
@@ -208,7 +212,7 @@
         for path, ie in inv.iter_entries():
             if ie.kind != 'file':
                 continue
-            if ie.text_version != rev_id:
+            if ie.revision != rev_id:
                 continue
             mutter('%s {%s} is changed in this revision',
                    path, ie.file_id)

=== modified file 'bzrlib/hashcache.py'
--- bzrlib/hashcache.py	2005-09-27 07:24:40 +0000
+++ bzrlib/hashcache.py	2005-09-30 02:54:51 +0000
@@ -30,13 +30,14 @@
 CACHE_HEADER = "### bzr hashcache v5\n"
 
 import os, stat, time
+import sha
 
 from bzrlib.osutils import sha_file
 from bzrlib.trace import mutter, warning
 from bzrlib.atomicfile import AtomicFile
 
 
-
+FP_MODE_COLUMN = 5
 
 def _fingerprint(abspath):
     try:
@@ -51,7 +52,7 @@
     # we discard any high precision because it's not reliable; perhaps we
     # could do better on some systems?
     return (fs.st_size, long(fs.st_mtime),
-            long(fs.st_ctime), fs.st_ino, fs.st_dev)
+            long(fs.st_ctime), fs.st_ino, fs.st_dev, fs.st_mode)
 
 
 class HashCache(object):
@@ -164,7 +165,16 @@
             return cache_sha1
         
         self.miss_count += 1
-        digest = sha_file(file(abspath, 'rb', buffering=65000))
+
+
+        mode = file_fp[FP_MODE_COLUMN]
+        if stat.S_ISREG(mode):
+            digest = sha_file(file(abspath, 'rb', buffering=65000))
+        elif stat.S_ISLNK(mode):
+            link_target = os.readlink(abspath)
+            digest = sha.new(os.readlink(abspath)).hexdigest()
+        else:
+            raise BzrError("file %r: unknown file stat mode: %o"%(abspath,mode))
 
         now = int(time.time())
         if file_fp[1] >= now or file_fp[2] >= now:
@@ -180,12 +190,8 @@
             self.update_count += 1
             self.needs_write = True
             self._cache[path] = (digest, file_fp)
-        
         return digest
         
-
-
-
     def write(self):
         """Write contents of cache to file."""
         outf = AtomicFile(self.cache_file_name(), 'wb')
@@ -244,7 +250,7 @@
 
             pos += 3
             fields = l[pos:].split(' ')
-            if len(fields) != 6:
+            if len(fields) != 7:
                 warning("bad line in hashcache: %r" % l)
                 continue
 

=== modified file 'bzrlib/inventory.py'
--- bzrlib/inventory.py	2005-09-29 20:34:25 +0000
+++ bzrlib/inventory.py	2005-10-03 01:42:16 +0000
@@ -36,7 +36,7 @@
 import bzrlib
 from bzrlib.errors import BzrError, BzrCheckError
 
-from bzrlib.osutils import quotefn, splitpath, joinpath, appendpath
+from bzrlib.osutils import quotefn, splitpath, joinpath, appendpath, sha_strings
 from bzrlib.trace import mutter
 from bzrlib.errors import NotVersionedError
 
@@ -53,14 +53,18 @@
         (within the parent directory)
 
     kind
-        'directory' or 'file'
+        'directory' or 'file' or 'symlink'
 
     parent_id
         file_id of the parent directory, or ROOT_ID
 
-    name_version
-        the revision_id in which the name or parent of this file was
-        last changed
+    revision
+        the revision_id in which this variation of this file was 
+        introduced.
+
+    executable
+        Indicates that this file should be executable on systems
+        that support it.
 
     text_sha1
         sha-1 of the text of the file
@@ -68,9 +72,6 @@
     text_size
         size in bytes of the text of the file
         
-    text_version
-        the revision_id in which the text of this file was introduced
-
     (reading a version 4 tree created a text_id field.)
 
     >>> i = Inventory()
@@ -115,9 +116,11 @@
     """
     
     __slots__ = ['text_sha1', 'text_size', 'file_id', 'name', 'kind',
-                 'text_id', 'parent_id', 'children',
-                 'text_version', 'name_version', ]
+                 'text_id', 'parent_id', 'children', 'executable', 
+                 'revision', 'symlink_target']
 
+    def _add_text_to_weave(self, new_lines, parents, weave_store):
+        weave_store.add_text(self.file_id, self.revision, new_lines, parents)
 
     def __init__(self, file_id, name, kind, parent_id, text_id=None):
         """Create an InventoryEntry
@@ -138,8 +141,8 @@
         if '/' in name or '\\' in name:
             raise BzrCheckError('InventoryEntry name %r is invalid' % name)
         
-        self.text_version = None
-        self.name_version = None
+        self.executable = False
+        self.revision = None
         self.text_sha1 = None
         self.text_size = None
         self.file_id = file_id
@@ -147,33 +150,90 @@
         self.kind = kind
         self.text_id = text_id
         self.parent_id = parent_id
+        self.symlink_target = None
         if kind == 'directory':
             self.children = {}
         elif kind == 'file':
             pass
+        elif kind == 'symlink':
+            pass
         else:
             raise BzrError("unhandled entry kind %r" % kind)
 
-
+    def read_symlink_target(self, path):
+        if self.kind == 'symlink':
+            try:
+                self.symlink_target = os.readlink(path)
+            except OSError,e:
+                raise BzrError("os.readlink error, %s" % e)
 
     def sorted_children(self):
         l = self.children.items()
         l.sort()
         return l
 
+    def check(self, checker, rev_id, inv, tree):
+        if self.parent_id != None:
+            if not inv.has_id(self.parent_id):
+                raise BzrCheckError('missing parent {%s} in inventory for revision {%s}'
+                        % (self.parent_id, rev_id))
+        if self.kind == 'file':
+            revision = self.revision
+            t = (self.file_id, revision)
+            if t in checker.checked_texts:
+                prev_sha = checker.checked_texts[t] 
+                if prev_sha != self.text_sha1:
+                    raise BzrCheckError('mismatched sha1 on {%s} in {%s}' %
+                                        (self.file_id, rev_id))
+                else:
+                    checker.repeated_text_cnt += 1
+                    return
+            mutter('check version {%s} of {%s}', rev_id, self.file_id)
+            file_lines = tree.get_file_lines(self.file_id)
+            checker.checked_text_cnt += 1 
+            if self.text_size != sum(map(len, file_lines)):
+                raise BzrCheckError('text {%s} wrong size' % self.text_id)
+            if self.text_sha1 != sha_strings(file_lines):
+                raise BzrCheckError('text {%s} wrong sha1' % self.text_id)
+            checker.checked_texts[t] = self.text_sha1
+        elif self.kind == 'directory':
+            if self.text_sha1 != None or self.text_size != None or self.text_id != None:
+                raise BzrCheckError('directory {%s} has text in revision {%s}'
+                        % (self.file_id, rev_id))
+        elif self.kind == 'root_directory':
+            pass
+        elif self.kind == 'symlink':
+            if self.text_sha1 != None or self.text_size != None or self.text_id != None:
+                raise BzrCheckError('symlink {%s} has text in revision {%s}'
+                        % (self.file_id, rev_id))
+            if self.symlink_target == None:
+                raise BzrCheckError('symlink {%s} has no target in revision {%s}'
+                        % (self.file_id, rev_id))
+        else:
+            raise BzrCheckError('unknown entry kind %r in revision {%s}' % 
+                                (self.kind, rev_id))
+
 
     def copy(self):
         other = InventoryEntry(self.file_id, self.name, self.kind,
                                self.parent_id)
+        other.executable = self.executable
         other.text_id = self.text_id
         other.text_sha1 = self.text_sha1
         other.text_size = self.text_size
-        other.text_version = self.text_version
-        other.name_version = self.name_version
+        other.symlink_target = self.symlink_target
+        other.revision = self.revision
         # note that children are *not* copied; they're pulled across when
         # others are added
         return other
 
+    def _get_snapshot_change(self, previous_entries):
+        if len(previous_entries) > 1:
+            return 'merged'
+        elif len(previous_entries) == 0:
+            return 'added'
+        else:
+            return 'modified/renamed/reparented'
 
     def __repr__(self):
         return ("%s(%r, %r, kind=%r, parent_id=%r)"
@@ -183,21 +243,63 @@
                    self.kind,
                    self.parent_id))
 
-    
+    def snapshot(self, revision, path, previous_entries, work_tree, 
+                 weave_store):
+        """Make a snapshot of this entry.
+        
+        This means that all its fields are populated, that it has its
+        text stored in the text store or weave.
+        """
+        mutter('new parents of %s are %r', path, previous_entries)
+        self._read_tree_state(path, work_tree)
+        if len(previous_entries) == 1:
+            # cannot be unchanged unless there is only one parent file rev.
+            parent_ie = previous_entries.values()[0]
+            if self._unchanged(path, parent_ie, work_tree):
+                mutter("found unchanged entry")
+                self.revision = parent_ie.revision
+                return "unchanged"
+        mutter('new revision for {%s}', self.file_id)
+        self.revision = revision
+        change = self._get_snapshot_change(previous_entries)
+        if self.kind != 'file':
+            return change
+        self._snapshot_text(previous_entries, work_tree, weave_store)
+        return change
+
+    def _snapshot_text(self, file_parents, work_tree, weave_store): 
+        mutter('storing file {%s} in revision {%s}',
+               self.file_id, self.revision)
+        # special case to avoid diffing on renames or 
+        # reparenting
+        if (len(file_parents) == 1
+            and self.text_sha1 == file_parents.values()[0].text_sha1
+            and self.text_size == file_parents.values()[0].text_size):
+            previous_ie = file_parents.values()[0]
+            weave_store.add_identical_text(
+                self.file_id, previous_ie.revision, 
+                self.revision, file_parents)
+        else:
+            new_lines = work_tree.get_file(self.file_id).readlines()
+            self._add_text_to_weave(new_lines, file_parents, weave_store)
+            self.text_sha1 = sha_strings(new_lines)
+            self.text_size = sum(map(len, new_lines))
+
     def __eq__(self, other):
         if not isinstance(other, InventoryEntry):
             return NotImplemented
 
-        return (self.file_id == other.file_id) \
-               and (self.name == other.name) \
-               and (self.text_sha1 == other.text_sha1) \
-               and (self.text_size == other.text_size) \
-               and (self.text_id == other.text_id) \
-               and (self.parent_id == other.parent_id) \
-               and (self.kind == other.kind) \
-               and (self.text_version == other.text_version) \
-               and (self.name_version == other.name_version)
-
+        return ((self.file_id == other.file_id)
+                and (self.name == other.name)
+                and (other.symlink_target == self.symlink_target)
+                and (self.text_sha1 == other.text_sha1)
+                and (self.text_size == other.text_size)
+                and (self.text_id == other.text_id)
+                and (self.parent_id == other.parent_id)
+                and (self.kind == other.kind)
+                and (self.revision == other.revision)
+                and (self.executable == other.executable)
+                )
 
     def __ne__(self, other):
         return not (self == other)
@@ -205,6 +307,32 @@
     def __hash__(self):
         raise ValueError('not hashable')
 
+    def _unchanged(self, path, previous_ie, work_tree):
+        compatible = True
+        # different inv parent
+        if previous_ie.parent_id != self.parent_id:
+            compatible = False
+        # renamed
+        elif previous_ie.name != self.name:
+            compatible = False
+        if self.kind == 'symlink':
+            if self.symlink_target != previous_ie.symlink_target:
+                compatible = False
+        if self.kind == 'file':
+            if self.text_sha1 != previous_ie.text_sha1:
+                compatible = False
+            else:
+                # FIXME: 20050930 probe for the text size when getting sha1
+                # in _read_tree_state
+                self.text_size = previous_ie.text_size
+        return compatible
+
+    def _read_tree_state(self, path, work_tree):
+        if self.kind == 'symlink':
+            self.read_symlink_target(work_tree.abspath(path))
+        if self.kind == 'file':
+            self.text_sha1 = work_tree.get_file_sha1(self.file_id)
+            self.executable = work_tree.is_executable(self.file_id)
 
 
 class RootEntry(InventoryEntry):

=== modified file 'bzrlib/merge.py'
--- bzrlib/merge.py	2005-09-30 06:39:42 +0000
+++ bzrlib/merge.py	2005-10-03 01:42:16 +0000
@@ -227,6 +227,7 @@
         self.tree = tree
         self.tempdir = tempdir
         os.mkdir(os.path.join(self.tempdir, "texts"))
+        os.mkdir(os.path.join(self.tempdir, "symlinks"))
         self.cached = {}
 
     def __iter__(self):
@@ -241,6 +242,9 @@
     def get_file_sha1(self, id):
         return self.tree.get_file_sha1(id)
 
+    def is_executable(self, id):
+        return self.tree.is_executable(id)
+
     def id2path(self, file_id):
         return self.tree.id2path(file_id)
 
@@ -263,13 +267,22 @@
         if self.root is not None:
             return self.tree.abspath(self.tree.id2path(id))
         else:
-            if self.tree.inventory[id].kind in ("directory", "root_directory"):
+            kind = self.tree.inventory[id].kind
+            if kind in ("directory", "root_directory"):
                 return self.tempdir
             if not self.cached.has_key(id):
-                path = os.path.join(self.tempdir, "texts", id)
-                outfile = file(path, "wb")
-                outfile.write(self.tree.get_file(id).read())
-                assert(os.path.exists(path))
+                if kind == "file":
+                    path = os.path.join(self.tempdir, "texts", id)
+                    outfile = file(path, "wb")
+                    outfile.write(self.tree.get_file(id).read())
+                    assert(bzrlib.osutils.lexists(path))
+                    if self.tree.is_executable(id):
+                        os.chmod(path, 0755)
+                else:
+                    assert kind == "symlink"
+                    path = os.path.join(self.tempdir, "symlinks", id)
+                    target = self.tree.get_symlink_target(id)
+                    os.symlink(target, path)
                 self.cached[id] = path
             return self.cached[id]
 

=== modified file 'bzrlib/merge_core.py'
--- bzrlib/merge_core.py	2005-09-30 06:39:42 +0000
+++ bzrlib/merge_core.py	2005-10-03 01:42:16 +0000
@@ -1,8 +1,10 @@
+import os.path
+
 import changeset
 from changeset import Inventory, apply_changeset, invert_dict
-import os.path
-from osutils import backup_file, rename
-from merge3 import Merge3
+from bzrlib.osutils import backup_file, rename
+from bzrlib.merge3 import Merge3
+import bzrlib
 
 class ApplyMerge3:
     """Contents-change wrapper around merge3.Merge3"""
@@ -206,6 +208,8 @@
     tree_entry = tree.tree.inventory[file_id]
     if tree_entry.kind == "file":
         return changeset.FileCreate(tree.get_file(file_id).read())
+    elif tree_entry.kind == "symlink":
+        return changeset.SymlinkCreate(tree.get_symlink_target(file_id))
     else:
         assert tree_entry.kind in ("root_directory", "directory")
         return changeset.dir_create
@@ -228,7 +232,7 @@
             return None
         if contents.new_contents is None:
             this_contents = get_contents(entry, this)
-            if this_path is not None and os.path.exists(this_path):
+            if this_path is not None and bzrlib.osutils.lexists(this_path):
                 if this_contents != contents.old_contents:
                     return conflict_handler.rem_contents_conflict(this_path, 
                         this_contents, contents.old_contents)
@@ -236,7 +240,7 @@
             else:
                 return None
         elif contents.old_contents is None:
-            if this_path is None or not os.path.exists(this_path):
+            if this_path is None or not bzrlib.osutils.lexists(this_path):
                 return contents
             else:
                 this_contents = get_contents(entry, this)
@@ -253,10 +257,18 @@
             raise Exception("Unhandled merge scenario")
 
 def make_merged_metadata(entry, base, other):
-    if entry.metadata_change is not None:
-        base_path = base.readonly_path(entry.id)
-        other_path = other.readonly_path(entry.id)    
-        return PermissionsMerge(base_path, other_path)
+    metadata = entry.metadata_change
+    if metadata is None:
+        return None
+    if isinstance(metadata, changeset.ChangeUnixPermissions):
+        if metadata.new_mode is None:
+            return None
+        elif metadata.old_mode is None:
+            return metadata
+        else:
+            base_path = base.readonly_path(entry.id)
+            other_path = other.readonly_path(entry.id)    
+            return PermissionsMerge(base_path, other_path)
     
 
 class PermissionsMerge(object):

=== removed file 'bzrlib/meta_store.py'
--- bzrlib/meta_store.py	2005-09-29 20:54:18 +0000
+++ bzrlib/meta_store.py	1970-01-01 00:00:00 +0000
@@ -1,67 +0,0 @@
-# Copyright (C) 2005 by Canonical Ltd
-
-# This program is free software; you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
-# the Free Software Foundation; either version 2 of the License, or
-# (at your option) any later version.
-
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
-
-# You should have received a copy of the GNU General Public License
-# along with this program; if not, write to the Free Software
-# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-
-from bzrlib.trace import mutter
-from bzrlib.store import Store
-from bzrlib.transport.local import LocalTransport
-
-class CachedStore(Store):
-    """A store that caches data locally, to avoid repeated downloads.
-    The precacache method should be used to avoid server round-trips for
-    every piece of data.
-    """
-    def __init__(self, store, cache_dir):
-        self.source_store = store
-        self.cache_store = store.__class__(LocalTransport(cache_dir))
-
-    def __getitem__(self, id):
-        mutter("Cache add %s" % id)
-        if id not in self.cache_store:
-            self.cache_store.add(self.source_store[id], id)
-        return self.cache_store[id]
-
-    def __contains__(self, fileid):
-        if fileid in self.cache_store:
-            return True
-        if fileid in self.source_store:
-            # We could copy at this time
-            return True
-        return False
-
-    def get(self, fileids, permit_failure=False, pb=None):
-        fileids = list(fileids)
-        hasids = self.cache_store.has(fileids)
-        needs = set()
-        for has, fileid in zip(hasids, fileids):
-            if not has:
-                needs.add(fileid)
-        if needs:
-            self.cache_store.copy_multi(self.source_store, needs,
-                    permit_failure=permit_failure)
-        return self.cache_store.get(fileids,
-                permit_failure=permit_failure, pb=pb)
-
-    def prefetch(self, ids):
-        """Copy a series of ids into the cache, before they are used.
-        For remote stores that support pipelining or async downloads, this can
-        increase speed considerably.
-
-        Failures while prefetching are ignored.
-        """
-        mutter("Prefetch of ids %s" % ",".join(ids))
-        self.cache_store.copy_multi(self.source_store, ids, 
-                                    permit_failure=True)
-

=== modified file 'bzrlib/osutils.py'
--- bzrlib/osutils.py	2005-09-28 09:35:50 +0000
+++ bzrlib/osutils.py	2005-09-30 02:54:51 +0000
@@ -95,6 +95,30 @@
     else:
         raise BzrError('invalid file kind %r' % kind)
 
+def lexists(f):
+    try:
+        if hasattr(os, 'lstat'):
+            os.lstat(f)
+        else:
+            os.stat(f)
+        return True
+    except OSError,e:
+        if e.errno == errno.ENOENT:
+            return False;
+        else:
+            raise BzrError("lstat/stat of (%r): %r" % (f, e))
+
+def normalizepath(f):
+    if hasattr(os.path, 'realpath'):
+        F = os.path.realpath
+    else:
+        F = os.path.abspath
+    [p,e] = os.path.split(f)
+    if e == "" or e == "." or e == "..":
+        return F(f)
+    else:
+        return os.path.join(F(p), e)
+    
 
 def backup_file(fn):
     """Copy a file to a backup.
@@ -141,6 +165,12 @@
     except OSError:
         return False
 
+def islink(f):
+    """True if f is a symlink."""
+    try:
+        return S_ISLNK(os.lstat(f)[ST_MODE])
+    except OSError:
+        return False
 
 def is_inside(dir, fname):
     """True if fname is inside dir.

=== modified file 'bzrlib/revision.py'
--- bzrlib/revision.py	2005-09-27 07:24:40 +0000
+++ bzrlib/revision.py	2005-10-02 22:47:02 +0000
@@ -32,28 +32,22 @@
     parent_ids
         List of parent revision_ids
     """
-    inventory_id = None
-    inventory_sha1 = None
-    revision_id = None
-    timestamp = None
-    message = None
-    timezone = None
-    committer = None
     
-    def __init__(self, **args):
+    def __init__(self, revision_id, **args):
+        self.revision_id = revision_id
         self.__dict__.update(args)
         self.parent_ids = []
         self.parent_sha1s = []
 
-
     def __repr__(self):
         return "<Revision id %s>" % self.revision_id
 
     def __eq__(self, other):
         if not isinstance(other, Revision):
             return False
-        return (self.inventory_id == other.inventory_id
-                and self.inventory_sha1 == other.inventory_sha1
+        # FIXME: rbc 20050930 parent_ids are not being compared
+        return (
+                self.inventory_sha1 == other.inventory_sha1
                 and self.revision_id == other.revision_id
                 and self.timestamp == other.timestamp
                 and self.message == other.message
@@ -64,7 +58,6 @@
         return not self.__eq__(other)
 
         
-
 REVISION_ID_RE = None
 
 def validate_revision_id(rid):
@@ -72,7 +65,7 @@
     global REVISION_ID_RE
     if not REVISION_ID_RE:
         import re
-        REVISION_ID_RE = re.compile('[\w.-]+@[\w.-]+--?\d+--?[0-9a-f]+\Z')
+        REVISION_ID_RE = re.compile('[\w:.-]+@[\w%.-]+--?[\w]+--?[0-9a-f]+\Z')
 
     if not REVISION_ID_RE.match(rid):
         raise ValueError("malformed revision-id %r" % rid)
@@ -211,6 +204,7 @@
     assert root not in ancestors[root]
     return root, ancestors, descendants
 
+
 def combined_graph(revision_a, revision_b, revision_source):
     """Produce a combined ancestry graph.
     Return graph root, ancestors map, descendants map, set of common nodes"""
@@ -232,6 +226,7 @@
         descendants[node].update(node_dec)
     return root, ancestors, descendants, common
 
+
 def common_ancestor(revision_a, revision_b, revision_source):
     try:
         root, ancestors, descendants, common = \
@@ -245,6 +240,7 @@
         raise bzrlib.errors.NoCommonAncestor(revision_a, revision_b)
     return farthest
 
+
 class MultipleRevisionSources(object):
     """Proxy that looks in multiple branches for revisions."""
     def __init__(self, *args):

=== modified file 'bzrlib/selftest/blackbox.py'
--- bzrlib/selftest/blackbox.py	2005-09-30 06:39:42 +0000
+++ bzrlib/selftest/blackbox.py	2005-10-02 21:51:29 +0000
@@ -30,6 +30,7 @@
 import os
 import shutil
 import sys
+import os
 
 from bzrlib.selftest import TestCaseInTempDir, BzrTestBase
 from bzrlib.branch import Branch
@@ -144,7 +145,7 @@
         file('goodbye', 'wt').write('baz')
         self.runbzr('add goodbye')
         self.runbzr('commit -m setup goodbye')
-        
+
         file('hello', 'wt').write('bar')
         file('goodbye', 'wt').write('qux')
         self.runbzr('revert hello')
@@ -159,6 +160,12 @@
         os.rmdir('revertdir')
         self.runbzr('revert')
 
+        os.symlink('/unlikely/to/exist', 'symlink')
+        self.runbzr('add symlink')
+        self.runbzr('commit -m f')
+        os.unlink('symlink')
+        self.runbzr('revert')
+        
         file('hello', 'wt').write('xyz')
         self.runbzr('commit -m xyz hello')
         self.runbzr('revert -r 1 hello')
@@ -402,6 +409,18 @@
         
 
 
+def has_symlinks():
+    if hasattr(os, 'symlink'):
+        return True
+    else:
+        return False
+
+def listdir_sorted(dir):
+    L = os.listdir(dir)
+    L.sort()
+    return L
+
+
 class OldTests(ExternalBase):
     """old tests moved from ./testbzr."""
 
@@ -572,3 +591,98 @@
 
         runbzr('info')
 
+        if has_symlinks():
+            progress("symlinks")
+            mkdir('symlinks')
+            chdir('symlinks')
+            runbzr('init')
+            os.symlink("NOWHERE1", "link1")
+            runbzr('add link1')
+            assert self.capture('unknowns') == ''
+            runbzr(['commit', '-m', '1: added symlink link1'])
+    
+            mkdir('d1')
+            runbzr('add d1')
+            assert self.capture('unknowns') == ''
+            os.symlink("NOWHERE2", "d1/link2")
+            assert self.capture('unknowns') == 'd1/link2\n'
+            # is d1/link2 found when adding d1
+            runbzr('add d1')
+            assert self.capture('unknowns') == ''
+            os.symlink("NOWHERE3", "d1/link3")
+            assert self.capture('unknowns') == 'd1/link3\n'
+            runbzr(['commit', '-m', '2: added dir, symlink'])
+    
+            runbzr('rename d1 d2')
+            runbzr('move d2/link2 .')
+            runbzr('move link1 d2')
+            assert os.readlink("./link2") == "NOWHERE2"
+            assert os.readlink("d2/link1") == "NOWHERE1"
+            runbzr('add d2/link3')
+            runbzr('diff')
+            runbzr(['commit', '-m', '3: rename of dir, move symlinks, add link3'])
+    
+            os.unlink("link2")
+            os.symlink("TARGET 2", "link2")
+            os.unlink("d2/link1")
+            os.symlink("TARGET 1", "d2/link1")
+            runbzr('diff')
+            assert self.capture("relpath d2/link1") == "d2/link1\n"
+            runbzr(['commit', '-m', '4: retarget of two links'])
+    
+            runbzr('remove d2/link1')
+            assert self.capture('unknowns') == 'd2/link1\n'
+            runbzr(['commit', '-m', '5: remove d2/link1'])
+    
+            os.mkdir("d1")
+            runbzr('add d1')
+            runbzr('rename d2/link3 d1/link3new')
+            assert self.capture('unknowns') == 'd2/link1\n'
+            runbzr(['commit', '-m', '6: remove d2/link1, move/rename link3'])
+            
+            runbzr(['check'])
+            
+            runbzr(['export', '-r', '1', 'exp1.tmp'])
+            chdir("exp1.tmp")
+            assert listdir_sorted(".") == [ "link1" ]
+            assert os.readlink("link1") == "NOWHERE1"
+            chdir("..")
+            
+            runbzr(['export', '-r', '2', 'exp2.tmp'])
+            chdir("exp2.tmp")
+            assert listdir_sorted(".") == [ "d1", "link1" ]
+            chdir("..")
+            
+            runbzr(['export', '-r', '3', 'exp3.tmp'])
+            chdir("exp3.tmp")
+            assert listdir_sorted(".") == [ "d2", "link2" ]
+            assert listdir_sorted("d2") == [ "link1", "link3" ]
+            assert os.readlink("d2/link1") == "NOWHERE1"
+            assert os.readlink("link2")    == "NOWHERE2"
+            chdir("..")
+            
+            runbzr(['export', '-r', '4', 'exp4.tmp'])
+            chdir("exp4.tmp")
+            assert listdir_sorted(".") == [ "d2", "link2" ]
+            assert os.readlink("d2/link1") == "TARGET 1"
+            assert os.readlink("link2")    == "TARGET 2"
+            assert listdir_sorted("d2") == [ "link1", "link3" ]
+            chdir("..")
+            
+            runbzr(['export', '-r', '5', 'exp5.tmp'])
+            chdir("exp5.tmp")
+            assert listdir_sorted(".") == [ "d2", "link2" ]
+            assert os.path.islink("link2")
+            assert listdir_sorted("d2")== [ "link3" ]
+            chdir("..")
+            
+            runbzr(['export', '-r', '6', 'exp6.tmp'])
+            chdir("exp6.tmp")
+            assert listdir_sorted(".") == [ "d1", "d2", "link2" ]
+            assert listdir_sorted("d1") == [ "link3new" ]
+            assert listdir_sorted("d2") == []
+            assert os.readlink("d1/link3new") == "NOWHERE3"
+            chdir("..")
+        else:
+            progress("skipping symlink tests")
+            

=== modified file 'bzrlib/selftest/test_commit.py'
--- bzrlib/selftest/test_commit.py	2005-09-27 07:24:40 +0000
+++ bzrlib/selftest/test_commit.py	2005-09-30 04:52:11 +0000
@@ -145,14 +145,14 @@
         self.assertFalse(tree1.has_filename('fruity'))
         self.check_inventory_shape(tree1.inventory, ['hello'])
         ie = tree1.inventory['hello-id']
-        eq(ie.name_version, 'test@rev-1')
+        eq(ie.revision, 'test@rev-1')
 
         tree2 = b.revision_tree('test@rev-2')
         eq(tree2.id2path('hello-id'), 'fruity')
         eq(tree2.get_file_text('hello-id'), 'contents of hello\n')
         self.check_inventory_shape(tree2.inventory, ['fruity'])
         ie = tree2.inventory['hello-id']
-        eq(ie.name_version, 'test@rev-2')
+        eq(ie.revision, 'test@rev-2')
 
 
     def test_reused_rev_id(self):
@@ -198,9 +198,9 @@
                                    ['a', 'a/b/hello', 'a/b'])
 
         inv = b.get_revision_inventory(r4)
-        eq(inv['hello-id'].name_version, r4)
-        eq(inv['a-id'].name_version, r1)
-        eq(inv['b-id'].name_version, r3)
+        eq(inv['hello-id'].revision, r4)
+        eq(inv['a-id'].revision, r1)
+        eq(inv['b-id'].revision, r3)
 
         
     def test_removed_commit(self):

=== modified file 'bzrlib/selftest/test_commit_merge.py'
--- bzrlib/selftest/test_commit_merge.py	2005-09-27 07:24:40 +0000
+++ bzrlib/selftest/test_commit_merge.py	2005-09-30 14:27:41 +0000
@@ -91,8 +91,8 @@
         commit(by, 'merge from x', rev_id='y@u-0-2', allow_pointless=False)
         tree = by.revision_tree('y@u-0-2')
         inv = tree.inventory
-        self.assertEquals(inv['ecks-id'].text_version, 'x@u-0-1')
-        self.assertEquals(inv['why-id'].text_version, 'y@u-0-1')
+        self.assertEquals(inv['ecks-id'].revision, 'x@u-0-1')
+        self.assertEquals(inv['why-id'].revision, 'y@u-0-1')
 
         check(bx)
         check(by)

=== modified file 'bzrlib/selftest/test_xml.py'
--- bzrlib/selftest/test_xml.py	2005-09-19 08:02:41 +0000
+++ bzrlib/selftest/test_xml.py	2005-09-30 14:27:41 +0000
@@ -60,16 +60,14 @@
 _committed_inv_v5 = """<inventory>
 <file file_id="bar-20050901064931-73b4b1138abc9cd2" 
       name="bar" parent_id="TREE_ROOT" 
-      text_version="mbp@foo-123123" name_version="mbp@foo-123123"
-      />
+      revision="mbp@foo-123123"/>
 <directory name="subdir"
            file_id="foo-20050801201819-4139aa4a272f4250"
            parent_id="TREE_ROOT" 
-           name_version="mbp@foo-00"/>
+           revision="mbp@foo-00"/>
 <file file_id="bar-20050824000535-6bc48cfad47ed134" 
       name="bar" parent_id="foo-20050801201819-4139aa4a272f4250" 
-      name_version="mbp@foo-00"
-      text_version="mbp@foo-123123"/>
+      revision="mbp@foo-00"/>
 </inventory>
 """
 
@@ -115,8 +113,7 @@
         eq(len(inv), 4)
         ie = inv['bar-20050824000535-6bc48cfad47ed134']
         eq(ie.kind, 'file')
-        eq(ie.text_version, 'mbp@foo-123123')
-        eq(ie.name_version, 'mbp@foo-00')
+        eq(ie.revision, 'mbp@foo-00')
         eq(ie.name, 'bar')
         eq(inv[ie.parent_id].kind, 'directory')
 

=== modified file 'bzrlib/selftest/testbranch.py'
--- bzrlib/selftest/testbranch.py	2005-09-30 07:38:34 +0000
+++ bzrlib/selftest/testbranch.py	2005-10-02 23:48:11 +0000
@@ -22,6 +22,7 @@
 from bzrlib.selftest import TestCaseInTempDir
 from bzrlib.trace import mutter
 
+
 class TestBranch(TestCaseInTempDir):
 
     def test_append_revisions(self):
@@ -32,9 +33,6 @@
         br.append_revision("rev2", "rev3")
         self.assertEquals(br.revision_history(), ["rev1", "rev2", "rev3"])
 
-
-class TestFetch(TestCaseInTempDir):
-
     def test_fetch_revisions(self):
         """Test fetch-revision operation."""
         from bzrlib.fetch import Fetcher
@@ -101,7 +99,47 @@
         self.assertTrue(os.path.exists('b/one'))
         self.assertFalse(os.path.exists('b/two'))
         
+
+    def test_record_initial_ghost_merge(self):
+        """A pending merge with no revision present is still a merge."""
+        branch = Branch.initialize('.')
+        branch.add_pending_merge('non:existent@rev--ision--0--2')
+        branch.commit('pretend to merge nonexistent-revision', rev_id='first')
+        rev = branch.get_revision(branch.last_revision())
+        self.assertEqual(len(rev.parent_ids), 1)
+        # parent_sha1s is not populated now, WTF. rbc 20051003
+        self.assertEqual(len(rev.parent_sha1s), 0)
+        self.assertEqual(rev.parent_ids[0], 'non:existent@rev--ision--0--2')
+
+# TODO 20051003 RBC:
+# compare the gpg-to-sign info for a commit with a ghost and 
+#     an identical tree without a ghost
+# fetch missing should rewrite the TOC of weaves to list newly available parents.
         
+    def test_pending_merges(self):
+        """Tracking pending-merged revisions."""
+        b = Branch.initialize('.')
+
+        self.assertEquals(b.pending_merges(), [])
+        b.add_pending_merge('foo@azkhazan-123123-abcabc')
+        self.assertEquals(b.pending_merges(), ['foo@azkhazan-123123-abcabc'])
+        b.add_pending_merge('foo@azkhazan-123123-abcabc')
+        self.assertEquals(b.pending_merges(), ['foo@azkhazan-123123-abcabc'])
+        b.add_pending_merge('wibble@fofof--20050401--1928390812')
+        self.assertEquals(b.pending_merges(),
+                          ['foo@azkhazan-123123-abcabc',
+                           'wibble@fofof--20050401--1928390812'])
+        b.commit("commit from base with two merges")
+        rev = b.get_revision(b.revision_history()[0])
+        self.assertEquals(len(rev.parent_ids), 2)
+        self.assertEquals(rev.parent_ids[0],
+                          'foo@azkhazan-123123-abcabc')
+        self.assertEquals(rev.parent_ids[1],
+                           'wibble@fofof--20050401--1928390812')
+        # list should be cleared when we do a commit
+        self.assertEquals(b.pending_merges(), [])
+ 
+
 # TODO: rewrite this as a regular unittest, without relying on the displayed output        
 #         >>> from bzrlib.commit import commit
 #         >>> bzrlib.trace.silent = True

=== modified file 'bzrlib/selftest/testfetch.py'
--- bzrlib/selftest/testfetch.py	2005-09-29 20:34:25 +0000
+++ bzrlib/selftest/testfetch.py	2005-10-02 23:48:11 +0000
@@ -54,15 +54,12 @@
     assert not has_revision(br_a, br_b.revision_history()[6])
     assert has_revision(br_a, br_b.revision_history()[5])
 
-    # When a non-branch ancestor is missing, it should be a failure, not
-    # exception
-    print ("CANNOT TEST MISSING NON REVISION_HISTORY ANCESTORS WITHOUT"
-           " GHOSTS")
-#    br_a4 = new_branch('br_a4')
-#    count, failures = greedy_fetch(br_a4, br_a)
-#    self.assertEqual(count, 6)
-#    self.assertEqual(failures, set((br_b.revision_history()[4],
-#                                    br_b.revision_history()[5]))) 
+    # When a non-branch ancestor is missing, it should be unlisted...
+    # as its not present in the ancestry weave.
+    br_b4 = new_branch('br_4')
+    count, failures = greedy_fetch(br_b4, br_b)
+    self.assertEqual(count, 7)
+    self.assertEqual(failures, [])
 
     self.assertEqual(greedy_fetch(writable_a, br_b)[0], 1)
     assert has_revision(br_a, br_b.revision_history()[3])

=== modified file 'bzrlib/selftest/testlog.py'
--- bzrlib/selftest/testlog.py	2005-09-29 20:34:25 +0000
+++ bzrlib/selftest/testlog.py	2005-10-02 21:51:29 +0000
@@ -65,6 +65,25 @@
             self.assertEquals(expected, got)
 
     def test_cur_revno(self):
+        b = Branch('.', init=True)
+
+        lf = LogCatcher()
+        b.commit('empty commit')
+        show_log(b, lf, verbose=True, start_revision=1, end_revision=1)
+        self.assertRaises(InvalidRevisionNumber, show_log, b, lf,
+                          start_revision=2, end_revision=1) 
+        self.assertRaises(InvalidRevisionNumber, show_log, b, lf,
+                          start_revision=1, end_revision=2) 
+        self.assertRaises(InvalidRevisionNumber, show_log, b, lf,
+                          start_revision=0, end_revision=2) 
+        self.assertRaises(InvalidRevisionNumber, show_log, b, lf,
+                          start_revision=1, end_revision=0) 
+        self.assertRaises(InvalidRevisionNumber, show_log, b, lf,
+                          start_revision=-1, end_revision=1) 
+        self.assertRaises(InvalidRevisionNumber, show_log, b, lf,
+                          start_revision=1, end_revision=-1) 
+
+    def test_cur_revno(self):
         b = Branch.initialize('.')
 
         lf = LogCatcher()

=== modified file 'bzrlib/selftest/testrevision.py'
--- bzrlib/selftest/testrevision.py	2005-09-29 20:34:25 +0000
+++ bzrlib/selftest/testrevision.py	2005-10-02 23:17:00 +0000
@@ -42,7 +42,7 @@
     b..3   a..3 merges b..4
     b..4   a..4
     b..5   a..5 merges b..5
-    b..6
+    b..6 merges a4
 
     so A is missing b6 at the start
     and B is missing a3, a4, a5
@@ -71,14 +71,10 @@
     fetch(from_branch=br2, to_branch=br1)
     br1.add_pending_merge(br2.revision_history()[5])
     commit(br1, "Commit nine", rev_id="a@u-0-5")
-    # disabled - it makes testing fetch too hard,
-    # but can be easily reenabled (without the fetch
-    # when GHOSTS are supported. RBC 20050928
+    # DO NOT FETCH HERE - we WANT a GHOST.
     #fetch(from_branch=br1, to_branch=br2)
-    #br2.add_pending_merge(br1.revision_history()[4])
-    commit(br2, "Commit ten - no merge", rev_id="b@u-0-6")
-
-    #fetch(from_branch=br2, to_branch=br1)
+    br2.add_pending_merge(br1.revision_history()[4])
+    commit(br2, "Commit ten - ghost merge", rev_id="b@u-0-6")
     
     return br1, br2
 
@@ -140,7 +136,6 @@
         ## assert not is_ancestor(revisions[3], revisions_2[3], br1)
 
 
-
 class TestIntermediateRevisions(TestCaseInTempDir):
 
     def setUp(self):
@@ -249,7 +244,7 @@
                           revisions_2[4])
         fetch(from_branch=br2, to_branch=br1)
         self.assertEqual(common_ancestor(revisions[5], revisions_2[6], sources),
-                          revisions_2[5])
+                          revisions[4]) # revisions_2[5] is equally valid
         self.assertEqual(common_ancestor(revisions_2[6], revisions[5], sources),
                           revisions_2[5])
 
@@ -260,7 +255,6 @@
         revisions = br1.revision_history()
         revisions_2 = br2.revision_history()
         sources = MultipleRevisionSources(br1, br2)
-
         expected_ancestors_list = {revisions[3]:(0, 0), 
                                    revisions[2]:(1, 1),
                                    revisions_2[4]:(2, 1), 
@@ -273,7 +267,6 @@
             self.assertEqual(ancestors_list[key], value, 
                               "key %r, %r != %r" % (key, ancestors_list[key],
                                                     value))
-
         self.assertEqual(common_ancestor(revisions[0], revisions[0], sources),
                           revisions[0])
         self.assertEqual(common_ancestor(revisions[1], revisions[2], sources),
@@ -287,9 +280,9 @@
         self.assertEqual(common_ancestor(revisions[4], revisions_2[5], sources),
                           revisions_2[4])
         self.assertEqual(common_ancestor(revisions[5], revisions_2[6], sources),
-                          revisions_2[5])
+                          revisions[4]) # revisions_2[5] is equally valid
         self.assertEqual(common_ancestor(revisions_2[6], revisions[5], sources),
-                          revisions_2[5])
+                          revisions[4]) # revisions_2[5] is equally valid
 
     def test_combined(self):
         """combined_graph

=== modified file 'bzrlib/selftest/teststatus.py'
--- bzrlib/selftest/teststatus.py	2005-09-29 20:34:25 +0000
+++ bzrlib/selftest/teststatus.py	2005-10-03 01:53:47 +0000
@@ -27,7 +27,7 @@
 class BranchStatus(TestCaseInTempDir):
     
     def test_branch_status(self): 
-        """Basic 'bzr mkdir' operation"""
+        """Test basic branch status"""
         from cStringIO import StringIO
         from bzrlib.status import show_status
         from bzrlib.branch import Branch
@@ -93,3 +93,48 @@
                            '  bye.c\n',
                            '  hello.c\n'])
 
+    def test_branch_status_specific_files(self): 
+        """Tests branch status with given specific files"""
+        from cStringIO import StringIO
+        from bzrlib.status import show_status
+        from bzrlib.branch import Branch
+        
+        b = Branch.initialize('.')
+
+        self.build_tree(['directory/','directory/hello.c', 'bye.c','test.c','dir2/'])
+        b.add('directory')
+        b.add('test.c')
+        b.commit('testing')
+        
+        tof = StringIO()
+        show_status(b, to_file=tof)
+        tof.seek(0)
+        self.assertEquals(tof.readlines(),
+                          ['unknown:\n',
+                           '  bye.c\n',
+                           '  dir2\n',
+                           '  directory/hello.c\n'
+                           ])
+
+        tof = StringIO()
+        show_status(b, specific_files=['bye.c','test.c','absent.c'], to_file=tof)
+        tof.seek(0)
+        self.assertEquals(tof.readlines(),
+                          ['unknown:\n',
+                           '  bye.c\n'
+                           ])
+        
+        tof = StringIO()
+        show_status(b, specific_files=['directory'], to_file=tof)
+        tof.seek(0)
+        self.assertEquals(tof.readlines(),
+                          ['unknown:\n',
+                           '  directory/hello.c\n'
+                           ])
+        tof = StringIO()
+        show_status(b, specific_files=['dir2'], to_file=tof)
+        tof.seek(0)
+        self.assertEquals(tof.readlines(),
+                          ['unknown:\n',
+                           '  dir2\n'
+                           ])

=== modified file 'bzrlib/selftest/teststore.py'
--- bzrlib/selftest/teststore.py	2005-09-29 21:13:03 +0000
+++ bzrlib/selftest/teststore.py	2005-10-02 21:51:29 +0000
@@ -16,7 +16,7 @@
 
 """Test Store implementation
 """
-from StringIO import StringIO
+from cStringIO import StringIO
 import os
 
 from bzrlib.store import copy_all
@@ -24,8 +24,9 @@
 from bzrlib.transport import NoSuchFile
 from bzrlib.store.compressed_text import CompressedTextStore
 from bzrlib.store.text import TextStore
-from bzrlib.selftest import TestCaseInTempDir
+from bzrlib.selftest import TestCase, TestCaseInTempDir
 from bzrlib.errors import BzrError, UnlistableStore
+import bzrlib.store
 
 
 def fill_store(store):
@@ -92,15 +93,19 @@
     check_equals(tester, store, ['b', 'd', 'c'], ['other', None, 'something'],
             permit_failure=True)
 
+
 def get_compressed_store(path='.'):
     t = LocalTransport(path)
     return CompressedTextStore(t)
 
+
 def get_text_store(path='.'):
     t = LocalTransport(path)
     return TextStore(t)
 
+
 class TestCompressedTextStore(TestCaseInTempDir):
+
     def test_multiple_add(self):
         """Multiple add with same ID should raise a BzrError"""
         store = get_compressed_store()
@@ -115,6 +120,13 @@
         test_ignore_get(self, store)
 
 
+    def test_total_size(self):
+        store = get_compressed_store('.')
+        store.add(StringIO('goodbye'), '123123')
+        store.add(StringIO('goodbye2'), '123123.dsc')
+        # these get gzipped - content should be stable
+        self.assertEqual(store.total_size(), (2, 55))
+        
     def test_copy_all(self):
         """Test copying"""
         os.mkdir('a')
@@ -125,6 +137,39 @@
         copy_all(store_a, store_b)
         self.assertEqual(store_a['1'].read(), 'foo')
         self.assertEqual(store_b['1'].read(), 'foo')
+
+class TestMemoryStore(TestCase):
+    
+    def get_store(self):
+        return bzrlib.store.ImmutableMemoryStore()
+    
+    def test_imports(self):
+        from bzrlib.store import ImmutableMemoryStore
+
+    def test_add_and_retrieve(self):
+        store = self.get_store()
+        store.add(StringIO('hello'), 'aa')
+        self.assertNotEqual(store['aa'], None)
+        self.assertEqual(store['aa'].read(), 'hello')
+        store.add(StringIO('hello world'), 'bb')
+        self.assertNotEqual(store['bb'], None)
+        self.assertEqual(store['bb'].read(), 'hello world')
+
+    def test_missing_is_absent(self):
+        store = self.get_store()
+        self.failIf('aa' in store)
+
+    def test_adding_fails_when_present(self):
+        store = self.get_store()
+        store.add(StringIO('hello'), 'aa')
+        self.assertRaises(bzrlib.store.StoreError,
+                          store.add, StringIO('hello'), 'aa')
+
+    def test_total_size(self):
+        store = self.get_store()
+        store.add(StringIO('goodbye'), '123123')
+        store.add(StringIO('goodbye2'), '123123.dsc')
+        self.assertEqual(store.total_size(), (2, 15))
         # TODO: Switch the exception form UnlistableStore to
         #       or make Stores throw UnlistableStore if their
         #       Transport doesn't support listing
@@ -161,4 +206,3 @@
         #       Transport doesn't support listing
         # store_c = RemoteStore('http://example.com/')
         # self.assertRaises(UnlistableStore, copy_all, store_c, store_b)
-

=== modified file 'bzrlib/selftest/whitebox.py'
--- bzrlib/selftest/whitebox.py	2005-09-29 20:34:25 +0000
+++ bzrlib/selftest/whitebox.py	2005-10-02 23:23:56 +0000
@@ -67,35 +67,6 @@
                           'Martin Pool <mbp@sourcefrog.net>-20050311061123-96a255005c7c9dbe')
 
 
-class PendingMerges(TestCaseInTempDir):
-
-    def test_pending_merges(self):
-        """Tracking pending-merged revisions."""
-        print "GHOST SUPPORT REMOVED"
-        return
-
-        b = Branch.initialize('.')
-
-        self.assertEquals(b.pending_merges(), [])
-        b.add_pending_merge('foo@azkhazan-123123-abcabc')
-        self.assertEquals(b.pending_merges(), ['foo@azkhazan-123123-abcabc'])
-        b.add_pending_merge('foo@azkhazan-123123-abcabc')
-        self.assertEquals(b.pending_merges(), ['foo@azkhazan-123123-abcabc'])
-        b.add_pending_merge('wibble@fofof--20050401--1928390812')
-        self.assertEquals(b.pending_merges(),
-                          ['foo@azkhazan-123123-abcabc',
-                           'wibble@fofof--20050401--1928390812'])
-        b.commit("commit from base with two merges")
-        rev = b.get_revision(b.revision_history()[0])
-        self.assertEquals(len(rev.parents), 2)
-        self.assertEquals(rev.parents[0].revision_id,
-                          'foo@azkhazan-123123-abcabc')
-        self.assertEquals(rev.parents[1].revision_id,
-                           'wibble@fofof--20050401--1928390812')
-        # list should be cleared when we do a commit
-        self.assertEquals(b.pending_merges(), [])
- 
-
 class MoreTests(TestCaseInTempDir):
 
     def test_revert(self):

=== modified file 'bzrlib/status.py'
--- bzrlib/status.py	2005-09-23 02:58:03 +0000
+++ bzrlib/status.py	2005-10-03 01:53:47 +0000
@@ -38,12 +38,14 @@
 
     show_pending
         If set, write pending merges.
+
     revision
         If None the compare latest revision with working tree
         If one revision show compared it with working tree.
         If two revisions show status between first and second.
     """
     import sys
+    from bzrlib.osutils import is_inside_any
     from bzrlib.delta import compare_trees
 
     if to_file == None:
@@ -83,11 +85,8 @@
             unknowns = new.unknowns()
             done_header = False
             for path in unknowns:
-                # FIXME: Should also match if the unknown file is within a
-                # specified directory.
-                if specific_files:
-                    if path not in specific_files:
-                        continue
+                if specific_files and not is_inside_any(specific_files, path):
+                    continue
                 if not done_header:
                     print >>to_file, 'unknown:'
                     done_header = True

=== modified file 'bzrlib/store/__init__.py'
--- bzrlib/store/__init__.py	2005-09-30 04:32:09 +0000
+++ bzrlib/store/__init__.py	2005-10-02 21:51:29 +0000
@@ -24,9 +24,12 @@
 unique ID.
 """
 
+from cStringIO import StringIO
+
 from bzrlib.errors import BzrError, UnlistableStore, TransportNotPossible
 from bzrlib.trace import mutter
 import bzrlib.transport
+from bzrlib.transport.local import LocalTransport
 
 ######################################################################
 # stores
@@ -34,23 +37,14 @@
 class StoreError(Exception):
     pass
 
+
 class Store(object):
     """This class represents the abstract storage layout for saving information.
+    
+    Files can be added, but not modified once they are in.  Typically
+    the hash is used as the name, or something else known to be unique,
+    such as a UUID.
     """
-    _transport = None
-    _max_buffered_requests = 10
-
-    def __init__(self, transport):
-        assert isinstance(transport, bzrlib.transport.Transport)
-        self._transport = transport
-
-    def __repr__(self):
-        if self._transport is None:
-            return "%s(None)" % (self.__class__.__name__)
-        else:
-            return "%s(%r)" % (self.__class__.__name__, self._transport.base)
-
-    __str__ = __repr__
 
     def __len__(self):
         raise NotImplementedError('Children should define their length')
@@ -187,6 +181,116 @@
         pb.clear()
         return len(to_copy), failed
 
+
+class TransportStore(Store):
+    """A TransportStore is a Store superclass for Stores that use Transports."""
+
+    _max_buffered_requests = 10
+
+    def __init__(self, transport):
+        assert isinstance(transport, bzrlib.transport.Transport)
+        super(TransportStore, self).__init__()
+        self._transport = transport
+
+    def __repr__(self):
+        if self._transport is None:
+            return "%s(None)" % (self.__class__.__name__)
+        else:
+            return "%s(%r)" % (self.__class__.__name__, self._transport.base)
+
+    __str__ = __repr__
+
+
+class ImmutableMemoryStore(Store):
+    """A memory only store."""
+
+    def __contains__(self, fileid):
+        return self._contents.has_key(fileid)
+
+    def __init__(self):
+        super(ImmutableMemoryStore, self).__init__()
+        self._contents = {}
+
+    def add(self, stream, fileid, compressed=True):
+        if self._contents.has_key(fileid):
+            raise StoreError("fileid %s already in the store" % fileid)
+        self._contents[fileid] = stream.read()
+
+    def __getitem__(self, fileid):
+        """Returns a file reading from a particular entry."""
+        if not self._contents.has_key(fileid):
+            raise IndexError
+        return StringIO(self._contents[fileid])
+
+    def _item_size(self, fileid):
+        return len(self._contents[fileid])
+
+    def __iter__(self):
+        return iter(self._contents.keys())
+
+    def total_size(self):
+        result = 0
+        count = 0
+        for fileid in self:
+            count += 1
+            result += self._item_size(fileid)
+        return count, result
+        
+
+class CachedStore(Store):
+    """A store that caches data locally, to avoid repeated downloads.
+    The precacache method should be used to avoid server round-trips for
+    every piece of data.
+    """
+
+    def __init__(self, store, cache_dir):
+        super(CachedStore, self).__init__()
+        self.source_store = store
+        # This clones the source store type with a locally bound
+        # transport. FIXME: it assumes a constructor is == cloning.
+        # clonable store - it might be nicer to actually have a clone()
+        # or something. RBC 20051003
+        self.cache_store = store.__class__(LocalTransport(cache_dir))
+
+    def __getitem__(self, id):
+        mutter("Cache add %s" % id)
+        if id not in self.cache_store:
+            self.cache_store.add(self.source_store[id], id)
+        return self.cache_store[id]
+
+    def __contains__(self, fileid):
+        if fileid in self.cache_store:
+            return True
+        if fileid in self.source_store:
+            # We could copy at this time
+            return True
+        return False
+
+    def get(self, fileids, permit_failure=False, pb=None):
+        fileids = list(fileids)
+        hasids = self.cache_store.has(fileids)
+        needs = set()
+        for has, fileid in zip(hasids, fileids):
+            if not has:
+                needs.add(fileid)
+        if needs:
+            self.cache_store.copy_multi(self.source_store, needs,
+                    permit_failure=permit_failure)
+        return self.cache_store.get(fileids,
+                permit_failure=permit_failure, pb=pb)
+
+    def prefetch(self, ids):
+        """Copy a series of ids into the cache, before they are used.
+        For remote stores that support pipelining or async downloads, this can
+        increase speed considerably.
+
+        Failures while prefetching are ignored.
+        """
+        mutter("Prefetch of ids %s" % ",".join(ids))
+        self.cache_store.copy_multi(self.source_store, ids, 
+                                    permit_failure=True)
+
+
 def copy_all(store_from, store_to):
     """Copy all ids from one store to another."""
     # TODO: Optional progress indicator

=== modified file 'bzrlib/store/compressed_text.py'
--- bzrlib/store/compressed_text.py	2005-09-29 21:03:18 +0000
+++ bzrlib/store/compressed_text.py	2005-10-02 21:51:29 +0000
@@ -30,7 +30,7 @@
 from StringIO import StringIO
 from stat import ST_SIZE
 
-class CompressedTextStore(bzrlib.store.Store):
+class CompressedTextStore(bzrlib.store.TransportStore):
     """Store that holds files indexed by unique names.
 
     Files can be added, but not modified once they are in.  Typically
@@ -56,8 +56,8 @@
     'goodbye'
     """
 
-    def __init__(self, basedir):
-        super(CompressedTextStore, self).__init__(basedir)
+    def __init__(self, transport):
+        super(CompressedTextStore, self).__init__(transport)
 
     def _check_fileid(self, fileid):
         if '\\' in fileid or '/' in fileid:
@@ -242,6 +242,7 @@
                 
         return count, total
 
+
 class ScratchCompressedTextStore(CompressedTextStore):
     """Self-destructing test subclass of CompressedTextStore.
 

=== modified file 'bzrlib/store/text.py'
--- bzrlib/store/text.py	2005-09-29 21:13:03 +0000
+++ bzrlib/store/text.py	2005-10-02 21:51:29 +0000
@@ -30,7 +30,8 @@
 from cStringIO import StringIO
 from stat import ST_SIZE
 
-class TextStore(bzrlib.store.Store):
+
+class TextStore(bzrlib.store.TransportStore):
     """Store that holds files indexed by unique names.
 
     Files can be added, but not modified once they are in.  Typically
@@ -38,11 +39,10 @@
     such as a UUID.
 
     Files are stored uncompressed, with no delta compression.
-
     """
 
-    def __init__(self, basedir):
-        super(TextStore, self).__init__(basedir)
+    def __init__(self, transport):
+        super(TextStore, self).__init__(transport)
 
     def _check_fileid(self, fileid):
         if not isinstance(fileid, basestring):
@@ -191,6 +191,7 @@
                 
         return count, total
 
+
 class ScratchTextStore(TextStore):
     """Self-destructing test subclass of TextStore.
 

=== modified file 'bzrlib/store/weave.py'
--- bzrlib/store/weave.py	2005-09-30 06:02:21 +0000
+++ bzrlib/store/weave.py	2005-10-02 21:51:29 +0000
@@ -115,6 +115,11 @@
         w.add(rev_id, parent_idxs, new_lines)
         self.put_weave(file_id, w)
         
+    def add_identical_text(self, file_id, old_rev_id, new_rev_id, parents):
+        w = self.get_weave_or_empty(file_id)
+        parent_idxs = map(w.lookup, parents)
+        w.add_identical(old_rev_id, new_rev_id, parent_idxs)
+        self.put_weave(file_id, w)
      
     def copy_multi(self, from_store, file_ids):
         assert isinstance(from_store, WeaveStore)

=== modified file 'bzrlib/transport/http.py'
--- bzrlib/transport/http.py	2005-09-29 21:03:18 +0000
+++ bzrlib/transport/http.py	2005-10-02 21:51:29 +0000
@@ -251,4 +251,3 @@
 
 register_transport('http://', HttpTransport)
 register_transport('https://', HttpTransport)
-

=== modified file 'bzrlib/transport/local.py'
--- bzrlib/transport/local.py	2005-09-18 21:09:19 +0000
+++ bzrlib/transport/local.py	2005-10-02 21:51:29 +0000
@@ -30,7 +30,10 @@
         """Set the base path where files will be stored."""
         if base.startswith('file://'):
             base = base[7:]
-        super(LocalTransport, self).__init__(os.path.realpath(base))
+        # realpath is incompatible with symlinks. When we traverse
+        # up we might be able to normpath stuff. RBC 20051003
+        super(LocalTransport, self).__init__(
+            os.path.normpath(os.path.abspath(base)))
 
     def should_cache(self):
         return False

=== modified file 'bzrlib/tree.py'
--- bzrlib/tree.py	2005-09-22 06:54:01 +0000
+++ bzrlib/tree.py	2005-10-03 01:42:16 +0000
@@ -134,7 +134,7 @@
     def get_file_lines(self, file_id):
         ie = self._inventory[file_id]
         weave = self.get_weave(file_id)
-        return weave.get(ie.text_version)
+        return weave.get(ie.revision)
         
 
     def get_file_text(self, file_id):
@@ -152,6 +152,9 @@
         if ie.kind == "file":
             return ie.text_sha1
 
+    def is_executable(self, file_id):
+        return self._inventory[file_id].executable
+
     def has_filename(self, filename):
         return bool(self.inventory.path2id(filename))
 
@@ -160,11 +163,17 @@
         for path, entry in self.inventory.iter_entries():
             yield path, 'V', entry.kind, entry.file_id
 
+    def get_symlink_target(self, file_id):
+        ie = self._inventory[file_id]
+        return ie.symlink_target;
 
 class EmptyTree(Tree):
     def __init__(self):
         self._inventory = Inventory()
 
+    def get_symlink_target(self, file_id):
+        return None
+
     def has_filename(self, filename):
         return False
 
@@ -274,6 +283,13 @@
             os.mkdir(fullpath)
         elif kind == 'file':
             pumpfile(tree.get_file(ie.file_id), file(fullpath, 'wb'))
+            if tree.is_executable(ie.file_id):
+                os.chmod(fullpath, 0755)
+        elif kind == 'symlink':
+            try:
+                os.symlink(ie.symlink_target, fullpath)
+            except OSError,e:
+                raise BzrError("Failed to create symlink %r -> %r, error: %s" % (fullpath, ie.symlink_target, e))
         else:
             raise BzrError("don't know how to export {%s} of kind %r" % (ie.file_id, kind))
         mutter("  export {%s} kind %s to %s" % (ie.file_id, kind, fullpath))
@@ -339,7 +355,10 @@
                 item.type = tarfile.REGTYPE
                 fileobj = tree.get_file(ie.file_id)
                 item.size = _find_file_size(fileobj)
-                item.mode = 0644
+                if tree.is_executable(ie.file_id):
+                    item.mode = 0755
+                else:
+                    item.mode = 0644
             else:
                 raise BzrError("don't know how to export {%s} of kind %r" %
                         (ie.file_id, ie.kind))

=== modified file 'bzrlib/upgrade.py'
--- bzrlib/upgrade.py	2005-09-30 05:56:05 +0000
+++ bzrlib/upgrade.py	2005-10-02 21:51:29 +0000
@@ -284,11 +284,9 @@
                 ie = inv[file_id]
                 if ie.kind == 'root_directory':
                     continue
-                assert hasattr(ie, 'name_version'), \
-                    'no name_version on {%s} in {%s}' % \
+                assert hasattr(ie, 'revision'), \
+                    'no revision on {%s} in {%s}' % \
                     (file_id, rev.revision_id)
-                if ie.kind == 'file':
-                    assert hasattr(ie, 'text_version')
 
         new_inv_xml = serializer_v5.write_inventory_to_string(inv)
         new_inv_sha1 = sha_string(new_inv_xml)
@@ -325,13 +323,13 @@
         parent_invs = map(self._load_updated_inventory, rev.parent_ids)
         for file_id in inv:
             ie = inv[file_id]
-            self._set_name_version(rev, ie, parent_invs)
+            self._set_revision(rev, ie, parent_invs)
             if ie.kind != 'file':
                 continue
             self._convert_file_version(rev, ie, parent_invs)
 
 
-    def _set_name_version(self, rev, ie, parent_invs):
+    def _set_revision(self, rev, ie, parent_invs):
         """Set name version for a file.
 
         Done in a slightly lazy way: if the file is renamed or in a merge revision
@@ -341,18 +339,18 @@
         if ie.kind == 'root_directory':
             return
         if len(parent_invs) != 1:
-            ie.name_version = rev.revision_id
+            ie.revision = rev.revision_id
         else:
             old_inv = parent_invs[0]
             if not old_inv.has_id(file_id):
-                ie.name_version = rev.revision_id
+                ie.revision = rev.revision_id
             else:
                 old_ie = old_inv[file_id]
                 if (old_ie.parent_id != ie.parent_id
                     or old_ie.name != ie.name):
-                    ie.name_version = rev.revision_id
+                    ie.revision = rev.revision_id
                 else:
-                    ie.name_version = old_ie.name_version
+                    ie.revision = old_ie.revision
 
 
 
@@ -373,10 +371,11 @@
         for parent_inv in parent_invs:
             if parent_inv.has_id(file_id):
                 parent_ie = parent_inv[file_id]
-                old_text_version = parent_ie.text_version
-                assert old_text_version in self.converted_revs 
-                if old_text_version not in file_parents:
-                    file_parents.append(old_text_version)
+                old_revision = parent_ie.revision
+                # if this fails, its a ghost ?
+                assert old_revision in self.converted_revs 
+                if old_revision not in file_parents:
+                    file_parents.append(old_revision)
                 if parent_ie.text_sha1 != ie.text_sha1:
                     text_changed = True
         if len(file_parents) != 1 or text_changed:
@@ -384,13 +383,13 @@
             assert sha_strings(file_lines) == ie.text_sha1
             assert sum(map(len, file_lines)) == ie.text_size
             w.add(rev_id, file_parents, file_lines, ie.text_sha1)
-            ie.text_version = rev_id
+            ie.revision = rev_id
             self.text_count += 1
             ##mutter('import text {%s} of {%s}',
             ##       ie.text_id, file_id)
         else:
             ##mutter('text of {%s} unchanged from parent', file_id)
-            ie.text_version = file_parents[0]
+            ie.revision = file_parents[0]
         del ie.text_id
 
 

=== modified file 'bzrlib/weave.py'
--- bzrlib/weave.py	2005-10-04 02:06:14 +0000
+++ bzrlib/weave.py	2005-10-04 02:27:27 +0000
@@ -350,6 +350,10 @@
 
         return new_version
 
+    def add_identical(self, old_rev_id, new_rev_id, parents):
+        """Add an identical text to old_rev_id as new_rev_id."""
+        old_lines = self.get(self.lookup(old_rev_id))
+        self.add(new_rev_id, parents, old_lines)
 
     def inclusions(self, versions):
         """Return set of all ancestors of given version(s)."""

=== modified file 'bzrlib/workingtree.py'
--- bzrlib/workingtree.py	2005-09-15 04:32:41 +0000
+++ bzrlib/workingtree.py	2005-10-03 01:42:16 +0000
@@ -21,6 +21,7 @@
 # it's not predictable when it will be written out.
 
 import os
+import stat
 import fnmatch
         
 import bzrlib.tree
@@ -69,7 +70,7 @@
         """
         inv = self._inventory
         for path, ie in inv.iter_entries():
-            if os.path.exists(self.abspath(path)):
+            if bzrlib.osutils.lexists(self.abspath(path)):
                 yield ie.file_id
 
 
@@ -83,7 +84,7 @@
         return os.path.join(self.basedir, filename)
 
     def has_filename(self, filename):
-        return os.path.exists(self.abspath(filename))
+        return bzrlib.osutils.lexists(self.abspath(filename))
 
     def get_file(self, file_id):
         return self.get_file_byname(self.id2path(file_id))
@@ -106,7 +107,7 @@
         if not inv.has_id(file_id):
             return False
         path = inv.id2path(file_id)
-        return os.path.exists(self.abspath(path))
+        return bzrlib.osutils.lexists(self.abspath(path))
 
 
     __contains__ = has_id
@@ -115,12 +116,22 @@
     def get_file_size(self, file_id):
         return os.path.getsize(self.id2abspath(file_id))
 
-
     def get_file_sha1(self, file_id):
         path = self._inventory.id2path(file_id)
         return self._hashcache.get_sha1(path)
 
 
+    def is_executable(self, file_id):
+        if os.name == "nt":
+            return self._inventory[file_id].executable
+        else:
+            path = self._inventory.id2path(file_id)
+            mode = os.lstat(self.abspath(path)).st_mode
+            return bool(stat.S_ISREG(mode) and stat.S_IEXEC&mode)
+
+    def get_symlink_target(self, file_id):
+        return os.readlink(self.id2path(file_id))
+
     def file_class(self, filename):
         if self.path2id(filename):
             return 'V'

=== modified file 'bzrlib/xml.py'
--- bzrlib/xml.py	2005-09-27 07:24:40 +0000
+++ bzrlib/xml.py	2005-09-30 02:54:51 +0000
@@ -1,3 +1,4 @@
+#! /usr/bin/env python
 # -*- coding: UTF-8 -*-
 
 # This program is free software; you can redistribute it and/or modify
@@ -72,4 +73,3 @@
 
     def _read_element(self, f):
         return ElementTree().parse(f)
-

=== modified file 'bzrlib/xml4.py'
--- bzrlib/xml4.py	2005-09-19 08:04:06 +0000
+++ bzrlib/xml4.py	2005-09-30 02:54:51 +0000
@@ -53,7 +53,7 @@
         if ie.text_size != None:
             e.set('text_size', '%d' % ie.text_size)
 
-        for f in ['text_id', 'text_sha1']:
+        for f in ['text_id', 'text_sha1', 'symlink_target']:
             v = getattr(ie, f)
             if v != None:
                 e.set(f, v)
@@ -100,6 +100,7 @@
                             parent_id)
         ie.text_id = elt.get('text_id')
         ie.text_sha1 = elt.get('text_sha1')
+        ie.symlink_target = elt.get('symlink_target')
 
         ## mutter("read inventoryentry: %r" % (elt.attrib))
 

=== modified file 'bzrlib/xml5.py'
--- bzrlib/xml5.py	2005-09-22 01:20:17 +0000
+++ bzrlib/xml5.py	2005-10-03 01:42:16 +0000
@@ -45,7 +45,7 @@
 
     def _pack_entry(self, ie):
         """Convert InventoryEntry to XML element"""
-        assert ie.kind == 'directory' or ie.kind == 'file'
+        assert ie.kind in ('directory', 'file', 'symlink')
         e = Element(ie.kind)
         e.set('name', ie.name)
         e.set('file_id', ie.file_id)
@@ -53,11 +53,14 @@
         if ie.text_size != None:
             e.set('text_size', '%d' % ie.text_size)
 
-        for f in ['text_version', 'text_sha1', 'name_version']:
+        for f in ['text_sha1', 'revision', 'symlink_target']:
             v = getattr(ie, f)
             if v != None:
                 e.set(f, v)
 
+        if ie.executable:
+            e.set('executable', 'yes')
+
         # to be conservative, we don't externalize the root pointers
         # for now, leaving them as null in the xml form.  in a future
         # version it will be implied by nested elements.
@@ -114,8 +117,8 @@
 
     def _unpack_entry(self, elt):
         kind = elt.tag
-        assert kind == 'directory' or kind == 'file', \
-	    'unsupported entry kind %s' % kind
+        if not kind in ('directory', 'file', 'symlink'):
+            raise AssertionError('unsupported entry kind %s' % kind)
 
         parent_id = elt.get('parent_id')
         if parent_id == None:
@@ -125,9 +128,11 @@
                             elt.get('name'),
                             kind,
                             parent_id)
-        ie.text_version = elt.get('text_version')
-        ie.name_version = elt.get('name_version')
+        ie.revision = elt.get('revision')
         ie.text_sha1 = elt.get('text_sha1')
+        ie.symlink_target = elt.get('symlink_target')
+        if elt.get('executable') == 'yes':
+            ie.executable = True
         v = elt.get('text_size')
         ie.text_size = v and int(v)
 

=== removed file 'patches/symlink-support.patch'
--- patches/symlink-support.patch	2005-05-17 07:01:47 +0000
+++ patches/symlink-support.patch	1970-01-01 00:00:00 +0000
@@ -1,629 +0,0 @@
-Return-Path: <erik@tntech.dk>
-X-Original-To: mbp@sourcefrog.net
-Delivered-To: mbp@ozlabs.org
-X-Greylist: delayed 1826 seconds by postgrey-1.21 at ozlabs; Sun, 15 May 2005 06:59:11 EST
-Received: from upstroke.tntech.dk (cpe.atm2-0-1041078.0x503eaf62.odnxx4.customer.tele.dk [80.62.175.98])
-	by ozlabs.org (Postfix) with ESMTP id B968E679EA
-	for <mbp@sourcefrog.net>; Sun, 15 May 2005 06:59:11 +1000 (EST)
-Received: by upstroke.tntech.dk (Postfix, from userid 1001)
-	id 63F83542FF; Sat, 14 May 2005 22:28:37 +0200 (CEST)
-To: Martin Pool <mbp@sourcefrog.net>
-Subject: [PATCH] symlink support patch
-From: Erik Toubro Nielsen <erik@tntech.dk>
-Date: Sat, 14 May 2005 22:28:37 +0200
-Message-ID: <86u0l57dsa.fsf@upstroke.tntech.dk>
-User-Agent: Gnus/5.1006 (Gnus v5.10.6) XEmacs/21.4 (Security Through
- Obscurity, linux)
-MIME-Version: 1.0
-Content-Type: multipart/mixed; boundary="=-=-="
-X-Spam-Checker-Version: SpamAssassin 3.0.3 (2005-04-27) on ozlabs.org
-X-Spam-Level: 
-X-Spam-Status: No, score=-1.4 required=3.2 tests=BAYES_00,NO_MORE_FUNN,
-	RCVD_IN_BLARS_RBL autolearn=no version=3.0.3
-Content-Length: 19688
-Lines: 604
-
---=-=-=
-
-I'm not sending this to the list as it is pretty large. 
-
-Let me know if its usefull and if I should rework anything.
-
-Erik
-
-
-Overview:
-
-        (Bugfix: in class TreeDelta I've moved kind= one level up, since
-        kind is also used in the else part)
-
-        Since both the InventoryEntry and stat cache is changed, perhaps
-        the branch format number should be increased?
-
-        Added test cases for symlinks to testbzr
-
-        Cannot use realpath since it expands path/L to path/LinkTarget
-        Cannot use exists, use new osutils.lexists
-
-        I'm overloading the text_modified to signify that a link        
-        target is changed. Perhaps text_modified should be renamed
-        content_modified?
-
-        InventoryEntry has a new member "symlink_target", 
-
-        The stat cache entry has been extended to contain the symlink
-        target and the st_mode. I try to ignore an old format cache
-        file.
-
-
-
---=-=-=
-Content-Type: text/x-patch
-Content-Disposition: inline; filename=symlinksupport.patch
-
-*** modified file 'bzrlib/add.py'
---- bzrlib/add.py 
-+++ bzrlib/add.py 
-@@ -38,7 +38,7 @@
-     for f in file_list:
-         kind = bzrlib.osutils.file_kind(f)
- 
--        if kind != 'file' and kind != 'directory':
-+        if kind != 'file' and kind != 'directory' and kind != 'symlink':
-             if f not in user_list:
-                 print "Skipping %s (can't add file of kind '%s')" % (f, kind)
-                 continue
-@@ -56,7 +56,7 @@
- 
-         kind = bzrlib.osutils.file_kind(f)
- 
--        if kind != 'file' and kind != 'directory':
-+        if kind != 'file' and kind != 'directory' and kind != 'symlink':
-             bailout("can't add file '%s' of kind %r" % (f, kind))
-             
-         versioned = (inv.path2id(rf) != None)
-
-*** modified file 'bzrlib/branch.py'
---- bzrlib/branch.py 
-+++ bzrlib/branch.py 
-@@ -58,11 +58,9 @@
-     run into the root."""
-     if f == None:
-         f = os.getcwd()
--    elif hasattr(os.path, 'realpath'):
--        f = os.path.realpath(f)
-     else:
--        f = os.path.abspath(f)
--    if not os.path.exists(f):
-+        f = bzrlib.osutils.normalizepath(f)
-+    if not bzrlib.osutils.lexists(f):
-         raise BzrError('%r does not exist' % f)
-         
- 
-@@ -189,7 +187,7 @@
-         """Return path relative to this branch of something inside it.
- 
-         Raises an error if path is not in this branch."""
--        rp = os.path.realpath(path)
-+        rp = bzrlib.osutils.normalizepath(path)
-         # FIXME: windows
-         if not rp.startswith(self.base):
-             bailout("path %r is not within branch %r" % (rp, self.base))
-@@ -531,7 +529,9 @@
-             file_id = entry.file_id
-             mutter('commit prep file %s, id %r ' % (p, file_id))
- 
--            if not os.path.exists(p):
-+            # it should be enough to use os.lexists instead of exists
-+            # but lexists in an 2.4 function
-+            if not bzrlib.osutils.lexists(p):
-                 mutter("    file is missing, removing from inventory")
-                 if verbose:
-                     show_status('D', entry.kind, quotefn(path))
-@@ -554,6 +554,10 @@
-             if entry.kind == 'directory':
-                 if not isdir(p):
-                     bailout("%s is entered as directory but not a directory" % quotefn(p))
-+            elif entry.kind == 'symlink':
-+                if not os.path.islink(p):
-+                    bailout("%s is entered as symbolic link but is not a symbolic link" % quotefn(p))
-+                entry.read_symlink_target(p)
-             elif entry.kind == 'file':
-                 if not isfile(p):
-                     bailout("%s is entered as file but is not a file" % quotefn(p))
-
-*** modified file 'bzrlib/diff.py'
---- bzrlib/diff.py 
-+++ bzrlib/diff.py 
-@@ -66,6 +66,11 @@
-         print >>to_file, "\\ No newline at end of file"
-     print >>to_file
- 
-+def _diff_symlink(old_tree, new_tree, file_id):
-+    t1 = old_tree.get_symlink_target(file_id)
-+    t2 = new_tree.get_symlink_target(file_id)
-+    print '*** *** target changed %r => %r' % (t1, t2)
-+    
- 
- def show_diff(b, revision, specific_files):
-     import sys
-@@ -112,12 +117,15 @@
- 
-     for old_path, new_path, file_id, kind, text_modified in delta.renamed:
-         print '*** renamed %s %r => %r' % (kind, old_path, new_path)
--        if text_modified:
-+        if kind == 'file' and text_modified:
-             _diff_one(old_tree.get_file(file_id).readlines(),
-                    new_tree.get_file(file_id).readlines(),
-                    sys.stdout,
-                    fromfile=old_label + old_path,
-                    tofile=new_label + new_path)
-+
-+        elif kind == 'symlink' and text_modified:
-+            _diff_symlink(old_tree, new_tree, file_id)
- 
-     for path, file_id, kind in delta.modified:
-         print '*** modified %s %r' % (kind, path)
-@@ -128,6 +136,8 @@
-                    fromfile=old_label + path,
-                    tofile=new_label + path)
- 
-+        elif kind == 'symlink':
-+            _diff_symlink(old_tree, new_tree, file_id)
- 
- 
- class TreeDelta:
-@@ -149,7 +159,9 @@
-     Each id is listed only once.
- 
-     Files that are both modified and renamed are listed only in
--    renamed, with the text_modified flag true.
-+    renamed, with the text_modified flag true. The text_modified
-+    applies either to the the content of the file or the target of the
-+    symbolic link, depending of the kind of file.
- 
-     The lists are normally sorted when the delta is created.
-     """
-@@ -224,8 +236,8 @@
-         specific_files = ImmutableSet(specific_files)
- 
-     for file_id in old_tree:
-+        kind = old_inv.get_file_kind(file_id)
-         if file_id in new_tree:
--            kind = old_inv.get_file_kind(file_id)
-             assert kind == new_inv.get_file_kind(file_id)
-             
-             assert kind in ('file', 'directory', 'symlink', 'root_directory'), \
-@@ -246,6 +258,14 @@
-                 old_sha1 = old_tree.get_file_sha1(file_id)
-                 new_sha1 = new_tree.get_file_sha1(file_id)
-                 text_modified = (old_sha1 != new_sha1)
-+            elif kind == 'symlink':
-+                t1 = old_tree.get_symlink_target(file_id)
-+                t2 = new_tree.get_symlink_target(file_id)
-+                if t1 != t2:
-+                    mutter("    symlink target changed")
-+                    text_modified = True
-+                else:
-+                    text_modified = False
-             else:
-                 ## mutter("no text to check for %r %r" % (file_id, kind))
-                 text_modified = False
-
-*** modified file 'bzrlib/inventory.py'
---- bzrlib/inventory.py 
-+++ bzrlib/inventory.py 
-@@ -125,14 +125,22 @@
-         self.kind = kind
-         self.text_id = text_id
-         self.parent_id = parent_id
-+        self.symlink_target = None
-         if kind == 'directory':
-             self.children = {}
-         elif kind == 'file':
-             pass
-+        elif kind == 'symlink':
-+            pass
-         else:
-             raise BzrError("unhandled entry kind %r" % kind)
- 
--
-+    def read_symlink_target(self, path):
-+        if self.kind == 'symlink':
-+            try:
-+                self.symlink_target = os.readlink(path)
-+            except OSError,e:
-+                raise BzrError("os.readlink error, %s" % e)
- 
-     def sorted_children(self):
-         l = self.children.items()
-@@ -145,6 +153,7 @@
-                                self.parent_id, text_id=self.text_id)
-         other.text_sha1 = self.text_sha1
-         other.text_size = self.text_size
-+        other.symlink_target = self.symlink_target
-         return other
- 
- 
-@@ -168,7 +177,7 @@
-         if self.text_size != None:
-             e.set('text_size', '%d' % self.text_size)
-             
--        for f in ['text_id', 'text_sha1']:
-+        for f in ['text_id', 'text_sha1', 'symlink_target']:
-             v = getattr(self, f)
-             if v != None:
-                 e.set(f, v)
-@@ -198,6 +207,7 @@
-         self = cls(elt.get('file_id'), elt.get('name'), elt.get('kind'), parent_id)
-         self.text_id = elt.get('text_id')
-         self.text_sha1 = elt.get('text_sha1')
-+        self.symlink_target = elt.get('symlink_target')
-         
-         ## mutter("read inventoryentry: %r" % (elt.attrib))
- 
-
-*** modified file 'bzrlib/osutils.py'
---- bzrlib/osutils.py 
-+++ bzrlib/osutils.py 
-@@ -58,7 +58,30 @@
-     else:
-         raise BzrError("can't handle file kind with mode %o of %r" % (mode, f)) 
- 
--
-+def lexists(f):
-+    try:
-+        if hasattr(os, 'lstat'):
-+            os.lstat(f)
-+        else:
-+            os.stat(f)
-+        return True
-+    except OSError,e:
-+        if e.errno == errno.ENOENT:
-+            return False;
-+        else:
-+            raise BzrError("lstat/stat of (%r): %r" % (f, e))
-+
-+def normalizepath(f):
-+    if hasattr(os.path, 'realpath'):
-+        F = os.path.realpath
-+    else:
-+        F = os.path.abspath
-+    [p,e] = os.path.split(f)
-+    if e == "" or e == "." or e == "..":
-+        return F(f)
-+    else:
-+        return os.path.join(F(p), e)
-+    
- 
- def isdir(f):
-     """True if f is an accessible directory."""
-
-*** modified file 'bzrlib/statcache.py'
---- bzrlib/statcache.py 
-+++ bzrlib/statcache.py 
-@@ -53,7 +53,7 @@
- to use a tdb instead.
- 
- The cache is represented as a map from file_id to a tuple of (file_id,
--sha1, path, size, mtime, ctime, ino, dev).
-+sha1, path, symlink target, size, mtime, ctime, ino, dev, mode).
- """
- 
- 
-@@ -62,11 +62,13 @@
- FP_CTIME = 2
- FP_INO   = 3
- FP_DEV   = 4
--
-+FP_ST_MODE=5
- 
- SC_FILE_ID = 0
- SC_SHA1    = 1 
--
-+SC_SYMLINK_TARGET = 3
-+
-+CACHE_ENTRY_SIZE = 10
- 
- def fingerprint(abspath):
-     try:
-@@ -79,7 +81,7 @@
-         return None
- 
-     return (fs.st_size, fs.st_mtime,
--            fs.st_ctime, fs.st_ino, fs.st_dev)
-+            fs.st_ctime, fs.st_ino, fs.st_dev, fs.st_mode)
- 
- 
- def _write_cache(basedir, entry_iter, dangerfiles):
-@@ -93,7 +95,9 @@
-                 continue
-             outf.write(entry[0] + ' ' + entry[1] + ' ')
-             outf.write(b2a_qp(entry[2], True))
--            outf.write(' %d %d %d %d %d\n' % entry[3:])
-+            outf.write(' ')
-+            outf.write(b2a_qp(entry[3], True)) # symlink_target
-+            outf.write(' %d %d %d %d %d %d\n' % entry[4:])
- 
-         outf.commit()
-     finally:
-@@ -114,10 +118,13 @@
-     
-     for l in cachefile:
-         f = l.split(' ')
-+        if len(f) != CACHE_ENTRY_SIZE:
-+            mutter("cache is in old format, must recreate it")
-+            return {}
-         file_id = f[0]
-         if file_id in cache:
-             raise BzrError("duplicated file_id in cache: {%s}" % file_id)
--        cache[file_id] = (f[0], f[1], a2b_qp(f[2])) + tuple([long(x) for x in f[3:]])
-+        cache[file_id] = (f[0], f[1], a2b_qp(f[2]), a2b_qp(f[3])) + tuple([long(x) for x in f[4:]])
-     return cache
- 
- 
-@@ -125,7 +132,7 @@
- 
- def _files_from_inventory(inv):
-     for path, ie in inv.iter_entries():
--        if ie.kind != 'file':
-+        if ie.kind != 'file' and ie.kind != 'symlink':
-             continue
-         yield ie.file_id, path
-     
-@@ -190,17 +197,24 @@
-         if (fp[FP_MTIME] >= now) or (fp[FP_CTIME] >= now):
-             dangerfiles.add(file_id)
- 
--        if cacheentry and (cacheentry[3:] == fp):
-+        if cacheentry and (cacheentry[4:] == fp):
-             continue                    # all stat fields unchanged
- 
-         hardcheck += 1
- 
--        dig = sha.new(file(abspath, 'rb').read()).hexdigest()
--
-+        mode = fp[FP_ST_MODE]
-+        if stat.S_ISREG(mode):
-+            link_target = '-' # can be anything, but must be non-empty
-+            dig = sha.new(file(abspath, 'rb').read()).hexdigest()
-+        elif stat.S_ISLNK(mode):
-+            link_target = os.readlink(abspath)
-+            dig = sha.new(link_target).hexdigest()
-+        else:
-+            raise BzrError("file %r: unknown file stat mode: %o"%(abspath,mode))
-         if cacheentry == None or dig != cacheentry[1]: 
-             # if there was no previous entry for this file, or if the
-             # SHA has changed, then update the cache
--            cacheentry = (file_id, dig, path) + fp
-+            cacheentry = (file_id, dig, path, link_target) + fp
-             cache[file_id] = cacheentry
-             change_cnt += 1
- 
-
-*** modified file 'bzrlib/tree.py'
---- bzrlib/tree.py 
-+++ bzrlib/tree.py 
-@@ -125,6 +125,11 @@
-                 os.mkdir(fullpath)
-             elif kind == 'file':
-                 pumpfile(self.get_file(ie.file_id), file(fullpath, 'wb'))
-+            elif kind == 'symlink':
-+                try:
-+                    os.symlink(ie.symlink_target, fullpath)
-+                except OSError,e:
-+                    bailout("Failed to create symlink %r -> %r, error: %s" % (fullpath, ie.symlink_target, e))
-             else:
-                 bailout("don't know how to export {%s} of kind %r" % (ie.file_id, kind))
-             mutter("  export {%s} kind %s to %s" % (ie.file_id, kind, fullpath))
-@@ -167,6 +172,9 @@
-         for path, entry in self.inventory.iter_entries():
-             yield path, 'V', entry.kind, entry.file_id
- 
-+    def get_symlink_target(self, file_id):
-+        ie = self._inventory[file_id]
-+        return ie.symlink_target;
- 
- class EmptyTree(Tree):
-     def __init__(self):
-@@ -179,7 +187,8 @@
-         if False:  # just to make it a generator
-             yield None
-     
--
-+    def get_symlink_target(self, file_id):
-+        return None
- 
- ######################################################################
- # diff
-@@ -245,3 +254,7 @@
-         if old_name != new_name:
-             yield (old_name, new_name)
-             
-+
-+    def get_symlink_target(self, file_id):
-+        ie = self._inventory[file_id]        
-+        return ie.symlink_target
-
-*** modified file 'bzrlib/workingtree.py'
---- bzrlib/workingtree.py 
-+++ bzrlib/workingtree.py 
-@@ -53,7 +53,7 @@
-             ie = inv[file_id]
-             if ie.kind == 'file':
-                 if ((file_id in self._statcache)
--                    or (os.path.exists(self.abspath(inv.id2path(file_id))))):
-+                    or (bzrlib.osutils.lexists(self.abspath(inv.id2path(file_id))))):
-                     yield file_id
- 
- 
-@@ -66,7 +66,7 @@
-         return os.path.join(self.basedir, filename)
- 
-     def has_filename(self, filename):
--        return os.path.exists(self.abspath(filename))
-+        return bzrlib.osutils.lexists(self.abspath(filename))
- 
-     def get_file(self, file_id):
-         return self.get_file_byname(self.id2path(file_id))
-@@ -86,7 +86,7 @@
-         self._update_statcache()
-         if file_id in self._statcache:
-             return True
--        return os.path.exists(self.abspath(self.id2path(file_id)))
-+        return bzrlib.osutils.lexists(self.abspath(self.id2path(file_id)))
- 
- 
-     __contains__ = has_id
-@@ -108,6 +108,12 @@
-         return self._statcache[file_id][statcache.SC_SHA1]
- 
- 
-+    def get_symlink_target(self, file_id):
-+        import statcache
-+        self._update_statcache()
-+        target = self._statcache[file_id][statcache.SC_SYMLINK_TARGET]
-+        return target
-+        
-     def file_class(self, filename):
-         if self.path2id(filename):
-             return 'V'
-
-*** modified file 'testbzr'
---- testbzr 
-+++ testbzr 
-@@ -1,4 +1,4 @@
--#! /usr/bin/python
-+#! /usr/bin/env python
- 
- # Copyright (C) 2005 Canonical Ltd
- 
-@@ -113,6 +113,17 @@
-     logfile.write('   at %s:%d\n' % stack[:2])
- 
- 
-+def has_symlinks():
-+    import os;
-+    if hasattr(os, 'symlink'):
-+        return True
-+    else:
-+        return False
-+
-+def listdir_sorted(dir):
-+    L = os.listdir(dir)
-+    L.sort()
-+    return L
- 
- # prepare an empty scratch directory
- if os.path.exists(TESTDIR):
-@@ -320,8 +331,105 @@
-     runcmd('bzr ignore *.blah')
-     assert backtick('bzr unknowns') == ''
-     assert file('.bzrignore', 'rt').read() == '*.blah\n'
--
--
-+    cd("..")
-+    
-+    if has_symlinks():
-+        progress("symlinks")
-+        mkdir('symlinks')
-+        cd('symlinks')
-+        runcmd('bzr init')
-+        os.symlink("NOWHERE1", "link1")
-+        runcmd('bzr add link1')
-+        assert backtick('bzr unknowns') == ''
-+        runcmd(['bzr', 'commit', '-m', '1: added symlink link1'])
-+
-+        mkdir('d1')
-+        runcmd('bzr add d1')
-+        assert backtick('bzr unknowns') == ''
-+        os.symlink("NOWHERE2", "d1/link2")
-+        assert backtick('bzr unknowns') == 'd1/link2\n'
-+        # is d1/link2 found when adding d1
-+        runcmd('bzr add d1')
-+        assert backtick('bzr unknowns') == ''
-+        os.symlink("NOWHERE3", "d1/link3")
-+        assert backtick('bzr unknowns') == 'd1/link3\n'
-+        runcmd(['bzr', 'commit', '-m', '2: added dir, symlink'])
-+
-+        runcmd('bzr rename d1 d2')
-+        runcmd('bzr move d2/link2 .')
-+        runcmd('bzr move link1 d2')
-+        assert os.readlink("./link2") == "NOWHERE2"
-+        assert os.readlink("d2/link1") == "NOWHERE1"
-+        runcmd('bzr add d2/link3')
-+        runcmd('bzr diff')
-+        runcmd(['bzr', 'commit', '-m', '3: rename of dir, move symlinks, add link3'])
-+
-+        os.unlink("link2")
-+        os.symlink("TARGET 2", "link2")
-+        os.unlink("d2/link1")
-+        os.symlink("TARGET 1", "d2/link1")
-+        runcmd('bzr diff')
-+        assert backtick("bzr relpath d2/link1") == "d2/link1\n"
-+        runcmd(['bzr', 'commit', '-m', '4: retarget of two links'])
-+
-+        runcmd('bzr remove d2/link1')
-+        assert backtick('bzr unknowns') == 'd2/link1\n'
-+        runcmd(['bzr', 'commit', '-m', '5: remove d2/link1'])
-+
-+        os.mkdir("d1")
-+        runcmd('bzr add d1')
-+        runcmd('bzr rename d2/link3 d1/link3new')
-+        assert backtick('bzr unknowns') == 'd2/link1\n'
-+        runcmd(['bzr', 'commit', '-m', '6: remove d2/link1, move/rename link3'])
-+        
-+        runcmd(['bzr', 'check'])
-+        
-+        runcmd(['bzr', 'export', '-r', '1', 'exp1.tmp'])
-+        cd("exp1.tmp")
-+        assert listdir_sorted(".") == [ "link1" ]
-+        assert os.readlink("link1") == "NOWHERE1"
-+        cd("..")
-+        
-+        runcmd(['bzr', 'export', '-r', '2', 'exp2.tmp'])
-+        cd("exp2.tmp")
-+        assert listdir_sorted(".") == [ "d1", "link1" ]
-+        cd("..")
-+        
-+        runcmd(['bzr', 'export', '-r', '3', 'exp3.tmp'])
-+        cd("exp3.tmp")
-+        assert listdir_sorted(".") == [ "d2", "link2" ]
-+        assert listdir_sorted("d2") == [ "link1", "link3" ]
-+        assert os.readlink("d2/link1") == "NOWHERE1"
-+        assert os.readlink("link2")    == "NOWHERE2"
-+        cd("..")
-+        
-+        runcmd(['bzr', 'export', '-r', '4', 'exp4.tmp'])
-+        cd("exp4.tmp")
-+        assert listdir_sorted(".") == [ "d2", "link2" ]
-+        assert os.readlink("d2/link1") == "TARGET 1"
-+        assert os.readlink("link2")    == "TARGET 2"
-+        assert listdir_sorted("d2") == [ "link1", "link3" ]
-+        cd("..")
-+        
-+        runcmd(['bzr', 'export', '-r', '5', 'exp5.tmp'])
-+        cd("exp5.tmp")
-+        assert listdir_sorted(".") == [ "d2", "link2" ]
-+        assert os.path.islink("link2")
-+        assert listdir_sorted("d2")== [ "link3" ]
-+        cd("..")
-+        
-+        runcmd(['bzr', 'export', '-r', '6', 'exp6.tmp'])
-+        cd("exp6.tmp")
-+        assert listdir_sorted(".") == [ "d1", "d2", "link2" ]
-+        assert listdir_sorted("d1") == [ "link3new" ]
-+        assert listdir_sorted("d2") == []
-+        assert os.readlink("d1/link3new") == "NOWHERE3"
-+        cd("..")
-+
-+        cd("..")
-+    else:
-+        progress("skipping symlink tests")
-+        
-     progress("all tests passed!")
- except Exception, e:
-     sys.stderr.write('*' * 50 + '\n'
-
-
---=-=-=--

